"""
KipetModel Mixins
"""
# Standard library imports
import copy

# Third party imports

# Kipet library imports
from kipet.core_methods.ParameterEstimator import wavelength_subset_selection


class WavelengthSelectionMixins():
    
    """Wrapper class mixin of wavelength subset selection methods for KipetModel"""
    
    def lack_of_fit(self):
        """Wrapper for ParameterEstimator lack_of_fit method"""
    
        lof = self.p_estimator.lack_of_fit()
        return lof
        
    def wavelength_correlation(self, corr_plot=False):
        """Wrapper for wavelength_correlation method in ParameterEstimator"""
        
        correlations = self.p_estimator.wavelength_correlation()
    
        if corr_plot:
            import matplotlib.pyplot as plt
            
            lists1 = sorted(correlations.items())
            x1, y1 = zip(*lists1)
            plt.plot(x1,y1)   
            plt.xlabel("Wavelength (cm)")
            plt.ylabel("Correlation between species and wavelength")
            plt.title("Correlation of species and wavelength")
            plt.show()   
    
        self.wavelength_correlations = correlations
        return correlations
    
    def run_lof_analysis(self, **kwargs):
        """Wrapper for run_lof_analysis method in ParameterEstimator"""
        
        builder_before_data = copy.copy(self.builder)
        builder_before_data.clear_data()
        
        end_time = self.settings.general.simulation_times[1]
        
        correlations = self.wavelength_correlation(corr_plot=False)
        lof = self.lack_of_fit()
        
        nfe = self.settings.collocation.nfe
        ncp = self.settings.collocation.ncp
        sigmas = self.settings.parameter_estimator.variances
        
        self.p_estimator.run_lof_analysis(builder_before_data, end_time, correlations, lof, nfe, ncp, sigmas, **kwargs)
    
        # Make a dict for the results - why is this not the case?
    
    def wavelength_subset_selection(self, n=0):
        """Wrapper for wavelength_subset_selection method in ParameterEstimator"""
        
        if n == 0:
            raise ValueError('You need to choose a subset!')
            
        if not hasattr(self, 'wavelength_correlations'):
            correlations = self.wavelength_correlation(corr_plot=False)
        else:
            correlations = self.wavelength_correlations
        
        return wavelength_subset_selection(correlations=correlations, n=n)
    
    def run_opt_with_subset_lambdas(self, wavelength_subset, **kwargs):
        """Wrapper for run_param_est_with_subset_lambdas method in ParameterEstimator"""
        
        solver = kwargs.pop('solver', 'k_aug')    
    
        builder_before_data = copy.copy(self.builder)
        builder_before_data.clear_data()
        
        #end_time = self.settings.general.simulation_times[1]
        end_time = self.p_estimator.model.end_time.value
        
        nfe = self.settings.collocation.nfe
        ncp = self.settings.collocation.ncp
        sigmas = self.settings.parameter_estimator.variances
        
        results = self.p_estimator.run_param_est_with_subset_lambdas(builder_before_data, end_time, wavelength_subset, nfe, ncp, sigmas, solver=solver)    
        return results
    


"""
Settings for KIPET
"""
# Standard library imports
import ast
import os
from pathlib import Path
import yaml

# Third party imports
import numpy as np

# Kipet library imports
from kipet.top_level.helper import AttrDict


class Settings():
    
    """This is a container for all of the options that can be used in Kipet
    Since it can be confusing due to the large number of options, this should
    make it easier for the user to see everything in one spot.
    
    """
    def __init__(self, category='model'):
        
        self._load_settings()
        
        # Initialize to the defaults (can be used at anytime)
        #if category == 'model':
        self.reset_model()
        # else:
        #     self.reset_block()
        
    def __str__(self):
        
        m = 25
          
        settings = 'Settings\n\n'
        
        if hasattr(self, 'general'):
            settings += 'General Settings:\n'
            for k, v in self.general.items():
                settings += f'{str(k).rjust(m)} : {v}\n'
        
        if hasattr(self, 'units'):
            settings += '\nUnit Settings:\n'
            for k, v in self.units.items():
                settings += f'{str(k).rjust(m)} : {v}\n'
        
        if hasattr(self, 'collocation'):
            settings += '\nCollocation Settings:\n'
            for k, v in self.collocation.items():
                settings += f'{str(k).rjust(m)} : {v}\n'
            
        if hasattr(self, 'simulator'):
            settings += '\nSimulation Settings:\n'
            for k, v in self.simulator.items():
                settings += f'{str(k).rjust(m)} : {v}\n'
            
        if hasattr(self, 'variance_estimator'):
            settings += '\nVarianceEstimator Settings:\n'
            for k, v in self.variance_estimator.items():
                settings += f'{str(k).rjust(m)} : {v}\n'
    
        if hasattr(self, 'parameter_estimator'):
            settings += '\nParameterEstimator Settings:\n'
            for k, v in self.parameter_estimator.items():
                settings += f'{str(k).rjust(m)} : {v}\n'
        
        if hasattr(self, 'solver'):
            settings += '\nSolver Settings:\n'
            for k, v in self.solver.items():
                settings += f'{str(k).rjust(m)} : {v}\n'
        
        return settings
        
    def __repr__(self):
        return self.__str__()
    
    @staticmethod
    def _get_file_path():
        
        current_dir = Path(__file__).parent
        settings_file = (current_dir / '../settings.yml').resolve()
        
        return settings_file
    
    def _load_settings(self):
        
        settings_file = self._get_file_path()
        
        with open(settings_file) as file:
            self.cfg = yaml.load(file, Loader=yaml.FullLoader)
    
            for sub_dict in self.cfg.values():
                for key, value in sub_dict.items():
                    if isinstance(value, bool):
                        continue
                    elif value in ['True', 'False', 'None']:
                        sub_dict[key] = ast.literal_eval(value)
                    elif is_number(value):
                        if float(value) == int(float(value)):
                            sub_dict[key] = int(float(value))
                        else:
                            sub_dict[key] = float(value)
                    else:
                        None
        
        return None
    
    def reset_model(self, specific_settings=None):
        """Initializes the settings dicts to their default values"""
        # Check which settings are absolutely necessary (really basic settings
        # and sepearate from the problem specific attributes)
        
        self.collocation = AttrDict(self.cfg['collocation'])
        self.simulator = AttrDict(self.cfg['simulator'])
        self.simulator.update({'solver_opts': AttrDict()})
        self.general = AttrDict(self.cfg['general'])
        self.variance_estimator = AttrDict(self.cfg['variance_estimator'])
        self.variance_estimator.update({'solver_opts': AttrDict()})
        self.parameter_estimator = AttrDict(self.cfg['parameter_estimator'])
        self.parameter_estimator.update({'solver_opts': AttrDict()})
        self.solver = AttrDict(self.cfg['solver'])
        self.units = AttrDict(self.cfg['units'])
        
        return None
    
    
    def update_settings(self, category, item, value):
        
        """Sets the default settings to some new value
        
        Don't break KIPET!
        """
        settings_file = self._get_file_path()
        
        self.cfg[category][item] = value
        with open(settings_file, 'w') as yaml_file:
            yaml_file.write( yaml.dump(self.cfg, default_flow_style=False))
    
        print(f'Updated default settings:\n\n{category}:\n\t{item}: {value}')
        return None
    
def is_number(s):
    """ Returns True if the input is a float (number), else False

    """
    try:
        float(s)
        return True
    except:
        return False
    
    
    
    



"""
FESimulator
"""
# Standard library imports

# Third party imports
from pyomo.environ import *
from pyomo.dae import *

# KIPET library imports
from kipet.core_methods.fe_factory import fe_initialize
from kipet.core_methods.PyomoSimulator import PyomoSimulator
from kipet.top_level.variable_names import VariableNames

__author__ = 'Michael Short, Kevin McBride'  #: July 2018 - January 2021

class FESimulator(PyomoSimulator):
    
    """This class is just an interface that allows the user of Kipet to easily
    implement the more general fe_factory class designed by David M. Thierry 
    without having to re-write the model to fit those arguments.
    """
    def __init__(self, model):
        """It takes in a standard Kipet/Pyomo model, rewrites it and calls 
        fe_factory. More information on fe_factory is included in that class
        description.
    
        Args:
            model (ConcreteModel): Pyomo model passed from KIPET
        
        """
        super(FESimulator, self).__init__(model)

        # KIPET variables
        self.__var = VariableNames()

        self.p_sim = PyomoSimulator(model)
        self.model = self.p_sim.model
        self.param_dict = {}
        self.param_name = self.__var.model_parameter
        
        # setattr(self.model, 'test_var', Var(self.model.alltime,
        #                                       [0],
        #                                       initialize=1.0))
        # print(self.model.test_var.display())

        self.c_sim = self.model.clone()
        
        # Check all parameters are fixed before simulating
        for param_obj in getattr(self.model, self.__var.model_parameter).values():
            if not param_obj.fixed:
                param_obj.fixed = True
        # Build the parameter dictionary in the format that fe_factory uses    
        model_var = self.__var.model_parameter
        
        for k, v in getattr(self.model, model_var).items():
            self.param_dict[model_var, k] = v.value

        #Build the initial condition dictionary in the format that fe_factory uses
        vars_to_init = [self.__var.concentration_model,
                        self.__var.state_model,
                        ]

        self.ics_ = {}
        for var in vars_to_init:
            self._var_init(var)        
    
        
    def _var_init(self, model_var):
        """Initializes the IC dict for the state variables and modifies it in
        place.
        
        Args:
            model_var (str): The model variable to be used in initializing the
            initial condition dict
            
        Returns:
            None
        """
        if hasattr(self.model, model_var):
            for k, v in getattr(self.model, model_var).items():
                st = self.model.start_time
                if k[0] == st:
                    self.ics_[model_var, k[1]] = v.value
                    
        return None
     
    def call_fe_factory(self, inputs_sub=None, dosing_points=None):    
        """This function applies all the inputs necessary for fe_factory to
        work, using Kipet syntax. Requires external inputs/dosing points to be
        specified with the following arguments.

        Args:
            inputs_sub (dict): dictionary of inputs 
            
            dosing_points (dict): dictionary of the dosing points
            
        Returns:
            None
        
        """
        self.inputs_sub=inputs_sub
        
        init = fe_initialize(self.model, 
                             self.c_sim,
                             init_con="init_conditions_c",
                             param_name=self.param_name,
                             param_values=self.param_dict,
                             inputs_sub=self.inputs_sub)
    
        init.load_initial_conditions(init_cond=self.ics_)

        if dosing_points is not None:
            init.load_discrete_jump(dosing_points)
            
        init.run()

        return None


"""
Helper functions for making KIPET models
"""
from kipet.top_level.variable_names import VariableNames

__var = VariableNames()

def step_fun(model, time_var, num='0', coeff=1, time=1e-3, fixed=True, switch=False, M=1, eta=1e-2, constant=0.5):
    """This formulates the step functions for KIPET using a sigmoidal function
    
    Args:
        model: The ConcreteModel using the step function
        
        time_var (float): Time point for step
        
        num (str): The name of the step function
        
        coeff (float): Coefficient for the step size
        
        time (float): The time for the step change (init if not fixed)
        
        fixed (bool): Choose if the time is known and fixed or variable
        
        on (bool): True if turning on, False if turning off, optional string args too
        
        M (float): Sigmoidal tuning parameter
        
        eta (float): Tuning parameter
        
        constant (float): Constant added to bring to certain value
        
    Returns:
        step_point (Expression): Returns the pyomo expression for the point
    
    """
    if switch in ['on', 'turn_on']:
        switch = True
    elif switch in ['off', 'turn_off']:
        switch = False
    
    factor = 1 if not switch else -1
    
    if fixed:
        if time is None:
            raise ValueError('Time cannot be None')
        time_var_step = factor*time
    else: 
        if time is not None:
            model.time_step_change[num].set_value(time)
        time_var_step = factor*model.time_step_change[num]
        
    pyomo_var = getattr(model, __var.dosing_variable)[time_var, __var.dosing_component]
    step_point = coeff*(0.5*((time_var_step - factor*pyomo_var) / ((time_var_step - factor*pyomo_var)**2 + eta**2/M) ** 0.5) + constant)
    
    return step_point


"""
Module to hold the method developed by Chen et al. 2016 based on Scipy
"""
import sys

import numpy as np
from pyomo.environ import *
from scipy.optimize import least_squares
from scipy.sparse import coo_matrix

from kipet.core_methods.data_tools import stdout_redirector
try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

def build_s_model(var_est_object):
    """Builds s_model to solve formulation 22 with ipopt

       This method is not intended to be used by users directly

    Args:

    Returns:
        None

    """
    var_est_object.S_model = ConcreteModel()
    if var_est_object._is_D_deriv:
        lower_bound = 0.0
    else:
        lower_bound = None
        
    var_est_object.S_model.S = Var(var_est_object._meas_lambdas,
                         var_est_object.component_set,
                         bounds=(lower_bound, None),
                         initialize=1.0)

    for l in var_est_object._meas_lambdas:
        for k in var_est_object.component_set:
            var_est_object.S_model.S[l, k].value = var_est_object.model.S[l, k].value
            if hasattr(var_est_object.model, 'known_absorbance'):
                if k in var_est_object.model.known_absorbance:
                    if var_est_object.model.S[l, k].value != var_est_object.model.known_absorbance_data[k][l]:
                        var_est_object.model.S[l, k].set_value(var_est_object.model.known_absorbance_data[k][l])
                        var_est_object.S_model.S[l, k].fix()
                        


def solve_s_scipy(var_est_object, **kwds):
    """Solves formulation 22 in weifengs paper (using scipy least_squares)

       This method is not intended to be used by users directly

    Args:
        tee (bool,optional): flag to tell the optimizer whether to stream output
        to the terminal or not
    
        profile_time (bool,optional): flag to tell pyomo to time the construction and solution of the model. 
        Default False

        ftol (float, optional): Tolerance for termination by the change of the cost function. Default is 1e-8

        xtol (float, optional): Tolerance for termination by the change of the independent variables. Default is 1e-8

        gtol (float, optional): Tolerance for termination by the norm of the gradient. Default is 1e-8.

        loss (str, optional): Determines the loss function. The following keyword values are allowed:
            'linear' (default) : rho(z) = z. Gives a standard least-squares problem.

            'soft_l1' : rho(z) = 2 * ((1 + z)**0.5 - 1). The smooth approximation of l1 (absolute value) loss. Usually a good choice for robust least squares.

            'huber' : rho(z) = z if z <= 1 else 2*z**0.5 - 1. Works similarly to 'soft_l1'.

            'cauchy' : rho(z) = ln(1 + z). Severely weakens outliers influence, but may cause difficulties in optimization process.

            'arctan' : rho(z) = arctan(z). Limits a maximum loss on a single residual, has properties similar to 'cauchy'.
        f_scale (float, optional): Value of soft margin between inlier and outlier residuals, default is 1.0

        max_nfev (int, optional): Maximum number of function evaluations before the termination

    Returns:
        None

    """
    method = kwds.pop('method','trf')
    def_tol = 1.4901161193847656e-07
    ftol = kwds.pop('ftol', def_tol)
    xtol = kwds.pop('xtol', def_tol)
    gtol = kwds.pop('gtol', def_tol)
    x_scale = kwds.pop('x_scale', 1.0)
    loss = kwds.pop('loss', 'linear')
    f_scale = kwds.pop('f_scale', 1.0)
    max_nfev = kwds.pop('max_nfev', None)
    verbose = kwds.pop('verbose', 2)
    profile_time = kwds.pop('profile_time', False)
    tee = kwds.pop('tee', False)

    if profile_time:
        print('-----------------Solve_S--------------------')
        t0 = time.time()

    if hasattr(var_est_object, '_abs_components'):
        n = var_est_object._nabs_components
    else:
        n = var_est_object._n_components

    for j, l in enumerate(var_est_object._meas_lambdas):
        for k, c in enumerate(var_est_object.component_set):
            if var_est_object.model.S[l, c].value < 0.0 and var_est_object._is_D_deriv == False:  #: only less thant zero for non-absorbing
                var_est_object._s_array[j * n + k] = 1e-2
            else:
                var_est_object._s_array[j * n + k] = var_est_object.model.S[l, c].value

    for j, t in enumerate(var_est_object._allmeas_times):
        for k, c in enumerate(var_est_object.component_set):
            var_est_object._z_array[j * n + k] = var_est_object.model.Z[t, c].value

    def F(x, z_array, d_array, nl, nt, nc):
        diff = np.zeros(nt*nl)
        for i in range(nt):
            for j in range(nl):
                diff[i*nl+j] = d_array[i, j]-sum(z_array[i*nc+k]*x[j*nc+k] for k in range(nc))
        return diff

    def JF(x, z_array, d_array, nl, nt, nc):
        row = []
        col = []
        data = []
        for i in range(nt):
            for j in range(nl):
                for k in range(nc):
                    row.append(i*nl+j)
                    col.append(j*nc+k)
                    data.append(-z_array[i*nc+k])
        return coo_matrix((data, (row, col)),
                          shape=(nt*nl, nc*nl))


    if var_est_object._is_D_deriv == False:
        lower_bound = 0.0
    else:
        lower_bound = -np.inf
            
    if tee:
        res = least_squares(F,
                            var_est_object._s_array,
                            JF,
                            (lower_bound ,np.inf),
                            method,
                            ftol,
                            xtol,
                            gtol,
                            x_scale,
                            loss,
                            f_scale,
                            max_nfev=max_nfev,
                            verbose=verbose,
                            args=(var_est_object._z_array,
                                  var_est_object._d_array,
                                  var_est_object._n_meas_lambdas,
                                  var_est_object._n_allmeas_times,
                                  n)
                            )
        
    else:
        f = StringIO()
        with stdout_redirector(f):
            res = least_squares(F,
                                var_est_object._s_array,
                                JF,
                                (lower_bound ,np.inf),
                                method,
                                ftol,
                                xtol,
                                gtol,
                                x_scale,
                                loss,
                                f_scale,
                                max_nfev=max_nfev,
                                verbose=verbose,
                                args=(var_est_object._z_array,
                                      var_est_object._d_array,
                                      var_est_object._n_meas_lambdas,
                                      var_est_object._n_allmeas_times,
                                      n)
                                )
            
        with open(var_est_object._tmp3,'w') as tf:
            tf.write(f.getvalue())
    
    if profile_time:
        t1 = time.time()
        print("Scipy.optimize.least_squares time={:.3f} seconds".format(t1-t0))

    for j, l in enumerate(var_est_object._meas_lambdas):
        for k, c in enumerate(var_est_object.component_set):
            var_est_object.model.S[l, c].value = res.x[j * n + k]
            if hasattr(var_est_object.model, 'known_absorbance'):
                if c in var_est_object.model.known_absorbance:
                    var_est_object.model.S[l, c].set_value(var_est_object.model.known_absorbance_data[c][l])

    return res.success

def build_c_model(var_est_object):
    """Builds s_model to solve formulation 25 with ipopt

       This method is not intended to be used by users directly

    Args:

    Returns:
        None

    """
    var_est_object.C_model = ConcreteModel()
    var_est_object.C_model.C = Var(var_est_object._allmeas_times,
                         var_est_object._sublist_components,
                         bounds=(0.0, None),
                         initialize=1.0)

    for l in var_est_object._allmeas_times:
        for k in var_est_object._sublist_components:
            var_est_object.C_model.C[l, k].value = var_est_object.model.C[l, k].value
            if hasattr(var_est_object.model, 'non_absorbing'):
                var_est_object.C_model.C[l, k].fix()

def solve_c_scipy(var_est_object, **kwds):
    """Solves formulation 25 in weifengs paper (using scipy least_squares)

       This method is not intended to be used by users directly

    Args:
        tee (bool,optional): flag to tell the optimizer whether to stream output
        to the terminal or not
    
        profile_time (bool,optional): flag to tell pyomo to time the construction and solution of the model. 
        Default False

        ftol (float, optional): Tolerance for termination by the change of the cost function. Default is 1e-8

        xtol (float, optional): Tolerance for termination by the change of the independent variables. Default is 1e-8

        gtol (float, optional): Tolerance for termination by the norm of the gradient. Default is 1e-8.

        loss (str, optional): Determines the loss function. The following keyword values are allowed:
            'linear' (default) : rho(z) = z. Gives a standard least-squares problem.

            'soft_l1' : rho(z) = 2 * ((1 + z)**0.5 - 1). The smooth approximation of l1 (absolute value) loss. Usually a good choice for robust least squares.

            'huber' : rho(z) = z if z <= 1 else 2*z**0.5 - 1. Works similarly to 'soft_l1'.

            'cauchy' : rho(z) = ln(1 + z). Severely weakens outliers influence, but may cause difficulties in optimization process.

            'arctan' : rho(z) = arctan(z). Limits a maximum loss on a single residual, has properties similar to 'cauchy'.
        f_scale (float, optional): Value of soft margin between inlier and outlier residuals, default is 1.0

        max_nfev (int, optional): Maximum number of function evaluations before the termination

    Returns:
        None

    """
    method = kwds.pop('method','trf')
    def_tol = 1.4901161193847656e-07
    ftol = kwds.pop('ftol',def_tol)
    xtol = kwds.pop('xtol',def_tol)
    gtol = kwds.pop('gtol',def_tol)
    x_scale = kwds.pop('x_scale',1.0)
    loss = kwds.pop('loss','linear')
    f_scale = kwds.pop('f_scale',1.0)
    max_nfev = kwds.pop('max_nfev',None)
    verbose = kwds.pop('verbose',2)
    profile_time = kwds.pop('profile_time',False)
    tee =  kwds.pop('tee',False)
    
    if profile_time:
        print('-----------------Solve_C--------------------')
        t0 = time.time()
   
    if hasattr(var_est_object, '_abs_components'):
        n = var_est_object._nabs_components
    else:
        n = var_est_object._n_components
        
    for i, t in enumerate(var_est_object._allmeas_times):
        for k, c in enumerate(var_est_object.component_set):
            if getattr(var_est_object.model, var_est_object.component_var)[t, c].value <= 0.0:
                var_est_object._c_array[i * n + k] = 1e-15
            else:
                var_est_object._c_array[i * n + k] = getattr(var_est_object.model, var_est_object.component_var)[t, c].value

    for j, l in enumerate(var_est_object._meas_lambdas):
        for k, c in enumerate(var_est_object.component_set):
            var_est_object._s_array[j * n + k] = var_est_object.model.S[l, c].value


    def F(x,s_array,d_array,nl,nt,nc):
        diff = np.zeros(nt*nl)
        for i in range(nt):
            for j in range(nl):
                diff[i*nl+j]=d_array[i,j]-sum(s_array[j*nc+k]*x[i*nc+k] for k in range(nc))
        return diff

    def JF(x,s_array,d_array,nl,nt,nc):
        row = []
        col = []
        data = []
        for i in range(nt):
            for j in range(nl):
                for k in range(nc):
                    row.append(i*nl+j)
                    col.append(i*nc+k)
                    data.append(-s_array[j*nc+k])

        return coo_matrix((data, (row, col)),
                          shape=(nt*nl,nc*nt))

    if tee:
        res = least_squares(F, 
                            var_est_object._c_array, 
                            JF,
                            (0.0, np.inf), 
                            method,
                            ftol, 
                            xtol, 
                            gtol,
                            x_scale, 
                            loss, 
                            f_scale,
                            max_nfev=max_nfev,
                            verbose=verbose,
                            args=(var_est_object._s_array,
                                  var_est_object._d_array,
                                  var_est_object._n_meas_lambdas,
                                  var_est_object._n_allmeas_times,
                                  n)
                            )
        
    else:
        f = StringIO()
        with stdout_redirector(f):
            res = least_squares(F, 
                                var_est_object._c_array, 
                                JF,
                                (0.0, np.inf), 
                                method,
                                ftol, 
                                xtol, 
                                gtol,
                                x_scale, 
                                loss, 
                                f_scale,
                                max_nfev=max_nfev,
                                verbose=verbose,
                                args=(var_est_object._s_array,
                                      var_est_object._d_array,
                                      var_est_object._n_meas_lambdas,
                                      var_est_object._n_allmeas_times,
                                      n)
                                )
                
        with open(var_est_object._tmp4, 'w') as tf:
            tf.write(f.getvalue())

    if profile_time:
        t1 = time.time()
        print("Scipy.optimize.least_squares time={:.3f} seconds".format(t1-t0))

    for j, t in enumerate(var_est_object._allmeas_times):
        for k,c in enumerate(var_est_object.component_set):
            getattr(var_est_object.model, var_est_object.component_var)[t,c].value = res.x[j*n+k]
   
    return res.success


# -*- coding: utf-8 -*-

from __future__ import print_function
import copy
import os
import re
import scipy.stats as st
import time

import matplotlib.pyplot as plt
import numpy as np
from pyomo import *
from pyomo.dae import *
from pyomo.environ import *
from pyomo.opt import (
    ProblemFormat,
    SolverFactory, 
    TerminationCondition,
)

from kipet.core_methods.Optimizer import *
from kipet.core_methods.TemplateBuilder import *
from kipet.common.read_hessian import *
from kipet.common.objectives import (
    conc_objective, 
    comp_objective,
    absorption_objective,
    )
from kipet.mixins.PEMixins import PEMixins 
from kipet.top_level.variable_names import VariableNames

class ParameterEstimator(PEMixins, Optimizer):
    """Optimizer for parameter estimation.

    Parameters
    ----------
    model : Pyomo model
        Pyomo model to be used in the parameter estimation

    """
    def __init__(self, model):
        super(ParameterEstimator, self).__init__(model)

        self.__var = VariableNames()

        self.hessian = None
        self._estimability = False
        self._idx_to_variable = dict()
        self._n_actual = self._n_components
        self.model_variance = True
        self.termination_condition = None
        
        # This should be a subclass or a mixin
        self.G_contribution = None
        self.unwanted_G = False
        self.time_variant_G = False
        self.time_invariant_G = False
        self.time_invariant_G_decompose = False
        self.time_invariant_G_no_decompose = False
        
        self.confidence_interval = 0.6826894921373

        if hasattr(self.model, 'non_absorbing'):
            warnings.warn("Overriden by non_absorbing")
            list_components = [k for k in self._mixture_components if k not in self._non_absorbing]
            self._sublist_components = list_components
            self._n_actual = len(self._sublist_components)

        # for new huplc structure (CS):
        if self._huplc_given:
            list_huplcabs = [k for k in self._huplc_absorbing]
            self._list_huplcabs = list_huplcabs
            self._n_huplc = len(list_huplcabs)

        if hasattr(self.model, 'known_absorbance'):
            warnings.warn("Overriden by known_absorbance")
            list_components = [k for k in self._mixture_components if k not in self._known_absorbance]
            self._sublist_components = list_components
            self._n_actual = len(self._sublist_components)

        else:
            self._sublist_components = [k for k in self._mixture_components]
            
        if hasattr(self, '_abs_components'):
            self.component_set = self._abs_components
            self.component_var = self.__var.concentration_spectra_abs
            self.n_val = self._nabs_components
            
        else:
            self.component_set = self._sublist_components
            self.component_var = self.__var.concentration_spectra
            self.n_val = self._n_actual
            
    def run_opt(self, solver, **kwds):

        """ Solves parameter estimation problem.

        Args:
            solver (str): name of the nonlinear solver to used

            solver_opts (dict, optional): options passed to the nonlinear solver

            variances (dict or float, optional): map of component name to noise variance. The
            map also contains the device noise variance. If not float then we only use device variance
            and ignore model variance.

            tee (bool,optional): flag to tell the optimizer whether to stream output
            to the terminal or not.

            with_d_vars (bool,optional): flag to the optimizer whether to add

            variables and constraints for D_bar(i,j).
            
            report_time (bool, optional): flag as to whether to time the parameter estimation or not.

            estimability (bool, optional): flag to tell the model whether it is
            being used by the estimability analysis and therefore will need to return the
            hessian for analysis.
            
            model_variance (bool, optional): Default is True. Flag to tell whether we are only
            considering the variance in the device, or also model noise as well.

            model_variance (bool, optional): Default is True. Flag to tell whether we are only
            considering the variance in the device, or also model noise as well.

        Returns:
            Results object with loaded results

        """
        run_opt_kwargs = copy.copy(kwds)
        
        solver_opts = kwds.pop('solver_opts', dict())
        variances = kwds.pop('variances', dict())
        tee = kwds.pop('tee', False)
        with_d_vars = kwds.pop('with_d_vars', False)
        covariance = kwds.pop('covariance', False)
        symbolic_solver_labels = kwds.pop('symbolic_solver_labels', False)

        estimability = kwds.pop('estimability', False)
        report_time = kwds.pop('report_time', False)
        model_variance = kwds.pop('model_variance', True)

        # additional arguments for inputs CS
        inputs = kwds.pop("inputs", None)
        inputs_sub = kwds.pop("inputs_sub", None)
        trajectories = kwds.pop("trajectories", None)
        fixedtraj = kwds.pop("fixedtraj", False)
        fixedy = kwds.pop("fixedy", False)
        yfix = kwds.pop("yfix", None)
        yfixtraj = kwds.pop("yfixtraj", None)

        jump = kwds.pop("jump", False)
        var_dic = kwds.pop("jump_states", None)
        jump_times = kwds.pop("jump_times", None)
        feed_times = kwds.pop("feed_times", None)
        
        confidence = kwds.pop('confidence_interval', None)
        if confidence is None:
            confidence = 0.6826894921373 # One standard deviation
        
        # user should input if the unwanted contribuiton is involved, and what type it is.
        # If it's time_invariant, St or Z_in should be inputed to check the rank of kernal of Omega matrix. KH.L
        G_contribution = kwds.pop('G_contribution', None)
        St = kwds.pop('St', dict())
        Z_in = kwds.pop('Z_in', dict())

        self.solver = solver
        self.model_variance = model_variance
        self._estimability = estimability
        self.confidence_interval = confidence

        if not self.model.alltime.get_discretization_info():
            raise RuntimeError('apply discretization first before initializing')

        self.G_contribution = G_contribution
       
        if report_time:
            start = time.time()

        opt = SolverFactory(self.solver)
        
        if self.G_contribution == 'time_invariant_G':
            self.decompose_G_test(St, Z_in)
        self.g_handling_status_messages()

        if covariance:
            if self.solver != 'ipopt_sens' and self.solver != 'k_aug':
                raise RuntimeError('To get covariance matrix the solver needs to be ipopt_sens or k_aug')
            
            if self.solver == 'ipopt_sens':
                if not 'compute_red_hessian' in solver_opts.keys():
                    solver_opts['compute_red_hessian'] = 'yes'
            
            if self.solver == 'k_aug':
                solver_opts['compute_inv'] = ''

            self._define_reduce_hess_order()

        if inputs_sub is not None:
            from kipet.common.additional_inputs import add_inputs
            
            add_kwargs = dict(
                fixedtraj = fixedtraj,
                fixedy = fixedy, 
                inputs_sub = inputs_sub,
                yfix = yfix,
                yfixtraj = yfixtraj,
                trajectories = trajectories,
            )
            
            add_inputs(self, add_kwargs)
        
        if jump:
            self.set_up_jumps(run_opt_kwargs)
            
        for key, val in solver_opts.items():
            opt.options[key] = val
            
        self.objective_value = 0
        active_objectives = [o for o in self.model.component_map(Objective, active=True)]        
        if active_objectives:
            print(
                "WARNING: The model has an active objective. Running optimization with models objective.\n"
                " To solve optimization with default objective (Weifengs) deactivate all objectives in the model.")
            solver_results = opt.solve(self.model, tee=tee)
    

        
        elif self._spectra_given:
            self.objective_value, self.cov_mat = self._solve_extended_model(variances, 
                                       opt,
                                       tee=tee,
                                       covariance=covariance,
                                       with_d_vars=with_d_vars,
                                       **kwds)
            
       
        elif self._concentration_given: # or self._custom_data_given:
            self.objective_value, self.cov_mat = self._solve_model_given_c(variances, 
                                      opt,
                                      tee=tee,
                                      covariance=covariance,
                                      **kwds)
       
        else:
            raise RuntimeError(
                'Must either provide concentration data or spectra in order to solve the parameter estimation problem')

        if report_time:
            end = time.time()
            print("Total execution time in seconds for variance estimation:", end - start)

        return self._get_results()


    def _get_results(self):
        """Removed results unit from function"""
    
        results = ResultsObject()
        
        results.objective = self.objective_value
        results.parameter_covariance = self.cov_mat
        results.load_from_pyomo_model(self.model)

        if self._spectra_given:
            self.compute_D_given_SC(results)

        if hasattr(self.model, self.__var.model_parameter_scaled): 
            setattr(results, self.__var.model_parameter, {name: getattr(self.model, self.__var.model_parameter)[name].value*getattr(self.model, self.__var.model_parameter_scaled)[name].value for name in self.model.parameter_names})
        else:
            setattr(results, self.__var.model_parameter, {name: getattr(self.model, self.__var.model_parameter)[name].value for name in self.model.parameter_names})

        if self.termination_condition!=None and self.termination_condition!=TerminationCondition.optimal:
            raise Exception("The current iteration was unsuccessful.")
        else:
            if self._estimability == True:
                return self.hessian, results
            else:
                return results

        return results

    def _get_list_components(self, species_list):
        
        if species_list is None:
            list_components = [k for k in self._mixture_components]
        else:
            list_components = []
            for k in species_list:
                if k in self._mixture_components:
                    list_components.append(k)
                else:
                    warnings.warn("Ignored {} since is not a mixture component of the model".format(k))

        return list_components
            
    def _get_objective_expr(self, model, with_d_vars, component_set, sigma_sq, device=False):
    
        expr = 0
        for t in model.meas_times:
            for l in model.meas_lambdas:
                if with_d_vars:
                    D_bar = model.D_bar[t, l]
                else:
                    if device:    
                        if hasattr(model, 'huplc_absorbing'):
                            D_bar = sum(model.Z[t, k] * model.S[l, k] for k in component_set if k not in model.solid_spec_arg1)
                        else:
                            D_bar = sum(model.Z[t, k] * model.S[l, k] for k in component_set)    
                    else:
                        if hasattr(model, '_abs_components'):
                            if hasattr(model, 'huplc_absorbing') and hasattr(model, 'solid_spec_arg1'):
                                D_bar = sum(model.Cs[t, k] * model.S[l, k] for k in model._abs_components if k not in model.solid_spec_arg1)
                            else:
                                D_bar = sum(model.Cs[t, k] * model.S[l, k] for k in model._abs_components)
                        else:
                            if hasattr(model, 'huplc_absorbing') and hasattr(model, 'solid_spec_arg1'):
                                D_bar = sum(model.C[t, k] * model.S[l, k] for k in component_set if k not in model.solid_spec_arg1)
                            else:
                                D_bar = sum(model.C[t, k] * model.S[l, k] for k in component_set)
                            
                if self.G_contribution == 'time_variant_G':
                    expr += (model.D[t, l] - D_bar - model.qr[t]*model.g[l]) ** 2 / (sigma_sq['device'])
                elif self.time_invariant_G_no_decompose:
                    expr += (model.D[t, l] - D_bar - model.g[l]) ** 2 / (sigma_sq['device'])
                else:
                    expr += (model.D[t, l] - D_bar) ** 2 / (sigma_sq['device'])
                   
        return expr 

    def _solve_extended_model(self, sigma_sq, optimizer, **kwds):
        """Solves estimation based on spectral data. (known variances)

           This method is not intended to be used by users directly
        Args:
            sigma_sq (dict): variances

            optimizer (SolverFactory): Pyomo Solver factory object

            tee (bool,optional): flag to tell the optimizer whether to stream output
            to the terminal or not

            with_d_vars (bool,optional): flag to the optimizer whether to add
            variables and constraints for D_bar(i,j)

            subset_lambdas (array_like,optional): Set of wavelengths to used in
            the optimization problem (not yet fully implemented). Default all wavelengths.

        Returns:
            None
        """
        tee = kwds.pop('tee', False)
        with_d_vars = kwds.pop('with_d_vars', False)

        if self._huplc_given:  # added for new huplc structure CS
            weights = kwds.pop('weights', [1.0, 1.0, 1.0])
        else:
            weights = kwds.pop('weights', [1.0, 1.0])
        warmstart = kwds.pop('warmstart', False)
        eigredhess2file = kwds.pop('eigredhess2file', False)
        penaltyparam = kwds.pop('penaltyparam', False)
        penaltyparamcon = kwds.pop('penaltyparamcon', False) #added for optional penalty term related to constraint CS
        ppenalty_dict = kwds.pop('ppenalty_dict', None)
        ppenalty_weights = kwds.pop('ppenalty_weights', None)
        covariance = kwds.pop('covariance', False)
        species_list = kwds.pop('subset_components', None)
        set_A = kwds.pop('subset_lambdas', list())
        self._eigredhess2file=eigredhess2file
        cov_mat = None

        if not set_A:
            set_A = self._meas_lambdas

        list_components = self._get_list_components(species_list)
        
        if not self._spectra_given:
            raise NotImplementedError("Extended model requires spectral data model.D[ti,lj]")

        if hasattr(self.model, 'non_absorbing'):
            warnings.warn("Overriden by non_absorbing!")
            list_components = [k for k in self._mixture_components if k not in self._non_absorbing]

        if hasattr(self.model, 'known_absorbance'):
            warnings.warn("Overriden by known_absorbance!")
            list_components = [k for k in self._mixture_components if k not in self._known_absorbance]

        all_sigma_specified = True

        if isinstance(sigma_sq, dict): 
            keys = sigma_sq.keys()
            for k in list_components:
                if k not in keys:
                    all_sigma_specified = False
                    sigma_sq[k] = max(sigma_sq.values())

            if not 'device' in sigma_sq.keys():
                all_sigma_specified = False
                sigma_sq['device'] = 1.0

        elif isinstance(sigma_sq, float):
            sigma_dev = sigma_sq
            sigma_sq = dict()
            sigma_sq['device'] = sigma_dev

        model = self.model
        
        # Instead of calling (19), (24) is called but all qr[i] are fixed at 1.0. KH.L
        if self.time_invariant_G_no_decompose:
            for i in model.alltime:
                model.qr[i] = 1.0
            model.qr.fix()

        def _qr_end_constraint(model):
            return model.qr[model.alltime[-1]] == 1.0
        
        if self.G_contribution == 'time_variant_G':
            model.qr_end_cons = Constraint(rule = _qr_end_constraint)
        
        
        if with_d_vars and self.model_variance:
            model.D_bar = Var(model.meas_times, model.meas_lambdas)
            
            def rule_D_bar(model, t, l):
                if hasattr(model, 'huplc_absorbing') and hasattr(model, 'solid_spec_arg1'):
                    return model.D_bar[t, l] == sum(
                        getattr(model, self.component_var)[t, k] * model.S[l, k] for k in self.component_set if k not in model.solid_spec_arg1)
                else:
                    return model.D_bar[t, l] == sum(getattr(model, self.component_var)[t, k] * model.S[l, k] for k in self.component_set)

            model.D_bar_constraint = Constraint(model.meas_times,
                                                model.meas_lambdas,
                                                rule=rule_D_bar)
        
        
        #For addition of huplc data and matching liquid and solid species (CS):
        def rule_Dhat_bar(model, t, l):
            list_huplcabs = [k for k in model.huplc_absorbing.value]
            return model.Dhat_bar[t, l] == (model.Z[t, l] + model.solidvol[t, l]) / (
                sum(model.Z[t, j] + model.solidvol[t, j] for j in list_huplcabs))

        def rule_objective(m):
            expr = self._get_objective_expr(m, with_d_vars, list_components, sigma_sq)

            expr *= weights[0]
            second_term = 0.0
            for t in m.meas_times:
                second_term += sum((model.C[t, k] - model.Z[t, k]) ** 2 / sigma_sq[k] for k in list_components)
                
            expr += weights[1] * second_term

            #for addition of L2 penalty term to objective penalizing values that deviate from values defined in ppenalty_dict (CS):
            if penaltyparam==True:
                if ppenalty_weights is None:
                    fourth_term = 0.0
                    for k in model.P.keys():
                        if k in ppenalty_dict.keys():
                            fourth_term = (model.P[k] - ppenalty_dict[k]) ** 2
                else:
                    if len(ppenalty_dict)!=len(ppenalty_weights):
                        raise RuntimeError(
                            'For every penalty term a weight must be defined.')
                    if ppenalty_dict.keys()!=ppenalty_weights.keys():
                        raise RuntimeError(
                            'Check the parameter names in ppenalty_weights and ppenalty_dict again. They must match.')
                    else:
                        fourth_term = 0.0
                        for k in model.P.keys():
                            if k in ppenalty_dict.keys():
                                fourth_term += ppenalty_weights[k] * (model.P[k] - ppenalty_dict[k]) ** 2
                expr +=fourth_term


            # for new huplc structure (CS):
            if hasattr(model, 'huplc_absorbing'):
                third_term = self._huplc_obj_term(model, sigma_sq)
                expr += weights[2] * third_term

                if penaltyparamcon == True: #added for optional penalty term related to constraint CS
                    rho=1e-1
                    sumpen=0.0
                    for t in model.alltime:
                        sumpen = sumpen + m.Y[t,'npen']
                    fifth_term = rho*sumpen
                    expr += fifth_term

            return expr

        

        # estimation without model variance and only device variance
        def rule_objective_device_only(model):
            
            if hasattr(model, 'huplc_absorbing'):
                component_set = model._abs_components
            else:
                component_set = list_components
            
            expr = self._get_objective_expr(model, with_d_vars, component_set, sigma_sq, device=True)
                
            if hasattr(model, 'huplc_absorbing'):
                third_term = self._huplc_obj_term(model, sigma_sq)
                expr += weights[2] * third_term
                      
            return expr

        if self.model_variance == True:
            model.objective = Objective(rule=rule_objective)
        else:
            model.objective = Objective(rule=rule_objective_device_only)

        #print(model.objective.expr.to_string())

        if warmstart==True:
            if hasattr(model,'dual') and hasattr(model,'ipopt_zL_out') and hasattr(model,'ipopt_zU_out') and hasattr(model,'ipopt_zL_in') and hasattr(model,'ipopt_zU_in'):
                self.update_warm_start(model)
            else:
                self.add_warm_start_suffixes(model, use_k_aug=False)

        if covariance and self.solver == 'ipopt_sens':
            hessian = self._covariance_ipopt_sens(model, optimizer, tee, all_sigma_specified)
            if self.model_variance:
                cov_mat = self._compute_covariance(hessian, sigma_sq)
            else:
                cov_mat = self._compute_covariance_no_model_variance(hessian, sigma_sq)
        
        elif covariance and self.solver == 'k_aug':
            hessian = self._covariance_k_aug(model, optimizer, tee, all_sigma_specified)
            if self.model_variance:
                cov_mat = self._compute_covariance(hessian, sigma_sq)
            else:
                cov_mat = self._compute_covariance_no_model_variance(hessian, sigma_sq)
        
        else:
            solver_results = optimizer.solve(model, tee=tee)
           
        if with_d_vars:
            model.del_component('D_bar')
            model.del_component('D_bar_constraint')
        
        obj_val = model.objective.expr()
        model.del_component('objective')
        
        return obj_val, cov_mat
        
    
    def _solve_model_given_c(self, sigma_sq, optimizer, **kwds):
        """Solves estimation based on concentration data. (known variances)

           This method is not intended to be used by users directly
        Args:
            sigma_sq (dict): variances

            optimizer (SolverFactory): Pyomo Solver factory object

            tee (bool,optional): flag to tell the optimizer whether to stream output
            to the terminal or not

        Returns:
            None
        """
        
        tee = kwds.pop('tee', False)
        if self._huplc_given:  # added for new huplc structure CS
            weights = kwds.pop('weights', [0.0, 1.0, 1.0])
        else:
            weights = kwds.pop('weights', [0.0, 1.0])
        covariance = kwds.pop('covariance', False)
        warmstart = kwds.pop('warmstart', False)
        eigredhess2file = kwds.pop('eigredhess2file', False)
        penaltyparam = kwds.pop('penaltyparam', False)
        penaltyparamcon = kwds.pop('penaltyparamcon', False) #added for optional penalty term related to constraint CS
        ppenalty_dict = kwds.pop('ppenalty_dict', None)
        ppenalty_weights = kwds.pop('ppenalty_weights', None)
        species_list = kwds.pop('subset_components', None)
        symbolic_solver_labels = kwds.pop('symbolic_solver_labels', False)
        cov_mat = None

        list_components = self._get_list_components(species_list)
        model = self.model

        if not self._concentration_given: # and not self._custom_data_given:
            raise NotImplementedError(
                "Parameter Estimation from concentration data requires concentration data model.C[ti,cj]")

        all_sigma_specified = True

        keys = sigma_sq.keys()
        for k in list_components:
            if hasattr(self, 'huplc_absorbing') and k not in keys:
                for ki in self.solid_spec_args1:
                    for key in keys:
                        if key == ki:
                            sigma_sq[ki] = sigma_sq[key]

            elif hasattr(self, 'huplc_absorbing') == False and k not in keys:
                all_sigma_specified = False
                sigma_sq[k] = max(sigma_sq.values())

        def rule_objective(model):
            obj=0
            if penaltyparamcon == True:
                rho = 100
                sumpen = 0.0
                obj += conc_objective(model, variance=sigma_sq)    
                obj += comp_objective(model)
                for t in model.allmeas_times:
                    sumpen += model.Y[t, 'npen']
                fifth_term = rho * sumpen
                obj += fifth_term
            else:
                obj += conc_objective(model, variance=sigma_sq)
                obj += comp_objective(model)
            return obj

        model.objective = Objective(rule=rule_objective)
        # print(model.objective.expr.to_string())

        if hasattr(model, 'custom_obj'):
            model.objective.expr += model.custom_obj
        
        #print(model.objective.expr.to_string())


        if warmstart==True:
            if hasattr(model,'dual') and hasattr(model,'ipopt_zL_out') and hasattr(model,'ipopt_zU_out') and hasattr(model,'ipopt_zL_in') and hasattr(model,'ipopt_zU_in'):
                self.update_warm_start(model)
            else:
                self.add_warm_start_suffixes(model)

        if covariance and self.solver == 'ipopt_sens':
            hessian = self._covariance_ipopt_sens(model, optimizer, tee, all_sigma_specified)
            if self._concentration_given:
                cov_mat = self._compute_covariance_C(hessian, sigma_sq)

        elif covariance and self.solver == 'k_aug':
            hessian = self._covariance_k_aug(model, optimizer, tee, all_sigma_specified, labels=True)
            if self._concentration_given:
                cov_mat = self._compute_covariance_C(hessian, sigma_sq)
            
        elif self.solver == 'gams' and covariance==False:
            ip = SolverFactory('gams')
            solver_results = ip.solve(model, solver='conopt', tee=True)
            
        else:
            solver_results = optimizer.solve(model, tee=tee, symbolic_solver_labels=True)
            self._termination_problems(solver_results, optimizer)
            
        obj_val = model.objective.expr()
        model.del_component('objective')
        
        return obj_val, cov_mat
    
    def _covariance_ipopt_sens(self, model, optimizer, tee, all_sigma_specified):
        """Generalize the covariance optimization with IPOPT Sens"""
        
        if self.model_variance == False:
            print("WARNING: FOR PROBLEMS WITH NO MODEL VARIANCE it is advised to use k_aug!!!")
        self._tmpfile = "ipopt_hess"
        solver_results = optimizer.solve(model, 
                                         tee=tee,
                                         logfile=self._tmpfile,
                                         report_timing=True)

        
        print("Done solving building reduce hessian")
        output_string = ''
        with open(self._tmpfile, 'r') as f:
            output_string = f.read()
        if os.path.exists(self._tmpfile):
            os.remove(self._tmpfile)

        ipopt_output, hessian_output = split_sipopt_string(output_string)
        
        if not all_sigma_specified:
            raise RuntimeError(
                'All variances must be specified to determine covariance matrix.\n Please pass variance dictionary to run_opt')

        n_vars = len(self._idx_to_variable)
        hessian = read_reduce_hessian(hessian_output, n_vars)
        
        return hessian
    
    def _covariance_k_aug(self, model, optimizer, tee, all_sigma_specified, labels=False):
        """Generalize the covariance optimization with k_aug"""
  
        self.add_warm_start_suffixes(model, use_k_aug=True)   
        
        count_vars = 1

        if self._spectra_given:
            for t in self._allmeas_times:
                for c in self.component_set:
                    getattr(model, self.component_var)[t, c].set_suffix_value(model.dof_v, count_vars)
                    count_vars += 1
        
            for l in self._meas_lambdas:
                for c in self.component_set:
                    model.S[l, c].set_suffix_value(model.dof_v, count_vars)
                    count_vars += 1

        for v in model.P.values():
            if v.is_fixed():
                continue
            model.P.set_suffix_value(model.dof_v, count_vars)
            count_vars += 1

        # Do the component initial values need to be considered in this matrix?

        if hasattr(model,'Pinit'):
            for k, v in self.model.Pinit.items():
            #for v in self.model.initparameter_names:
                model.init_conditions[k].set_suffix_value(model.dof_v, count_vars)
                count_vars += 1

        self._tmpfile = "k_aug_hess"
        ip = SolverFactory('ipopt')
        solver_results = ip.solve(model,
                                  tee=tee,
                                  logfile=self._tmpfile,
                                  report_timing=True,
                                  symbolic_solver_labels=labels,
                                  )
        
         
        if labels:
            self._termination_problems(solver_results, optimizer)
        
        k_aug = SolverFactory('k_aug')
        
        model.ipopt_zL_in.display()
        model.rh_name.display()
        self.update_warm_start(model)
        model.ipopt_zL_in.display()
        
        k_aug.solve(model, tee=tee)
        print("Done solving building reduce hessian")

        if not all_sigma_specified:
            raise RuntimeError(
                'All variances must be specified to determine covariance matrix.\n Please pass variance dictionary to run_opt')

        n_vars = len(self._idx_to_variable)         
        var_loc = model.rh_name
        for v in self._idx_to_variable.values():
            try:
                var_loc[v]
            except:
                var_loc[v] = 0

        vlocsize = len(var_loc)
        unordered_hessian = np.loadtxt('result_red_hess.txt')
        if os.path.exists('result_red_hess.txt'):
            os.remove('result_red_hess.txt')
        
        hessian = self._order_k_aug_hessian(unordered_hessian, var_loc)
        #hessian =self..order_k_aug_hessian(self, unordered_hessian, var_loc)
        
        if self._estimability == True:
            self.hessian = hessian

        return hessian
    
    def _termination_problems(self, solver_results, optimizer):
        """This is some funky code - do we need it?"""
        
        self.termination_condition = solver_results.solver.termination_condition
        if self.termination_condition != TerminationCondition.optimal:
            print("WARNING: The solution of the iteration was unsuccessful. The problem is solved with additional solver options.")
            optimizer.options["OF_start_with_resto"] = 'yes'
            solver_results = optimizer.solve(m, tee=tee, symbolic_solver_labels=True)
            self.termination_condition = solver_results.solver.termination_condition
            if self.termination_condition != TerminationCondition.optimal:
                print(
                    "WARNING: The solution of the iteration was unsuccessful. The problem is solved with additional solver options.")
                optimizer.options["OF_start_with_resto"] = 'no'
                # optimizer.options["OF_bound_push"] = 1E-02
                optimizer.options["OF_bound_relax_factor"] = 1E-05
                solver_results = optimizer.solve(m, tee=tee, symbolic_solver_labels=True)
                self.termination_condition = solver_results.solver.termination_condition
                # options["OF_bound_relax_factor"] = 1E-08
                if self.termination_condition != TerminationCondition.optimal:
                        print("The current iteration was unsuccessful.")
                        
        return None

    @staticmethod
    def _huplc_obj_term(m, sigma_sq):
        third_term = 0.0
        if not 'device-huplc' in sigma_sq.keys():
            sigma_sq['device-huplc'] = 1.0
        if hasattr(m, 'solid_spec_arg1') and hasattr(m, 'solid_spec_arg2'):
            solidvol_dict = dict()
            m.add_component('cons_solidvol', ConstraintList())
            new_consolidvol = getattr(m, 'cons_solidvol')
            m.add_component('solidvol', Var(m.huplctime, m.huplc_absorbing.value, initialize=solidvol_dict))

            for k in m.solid_spec_arg1:
                for j in m.algebraics:
                    for time in m.huplctime:
                        if j == k:
                            for l in m.huplc_absorbing.value:
                                strsolidspec = "\'" + str(k) + "\'" + ', ' + "\'" + str(
                                    l) + "\'"  # pair of absorbing solid and liquid
                                if l in m.solid_spec_arg2 and strsolidspec in str(
                                        m.solid_spec.keys()):  #check whether pair of solid and liquid in keys and whether liquid in huplcabs species
                                    valY = value(m.Y[time, k]) / value(m.vol)
                                    if valY <= 0:
                                        new_consolidvol.add(m.solidvol[time, l] == 0.0)
                                    else:
                                        new_consolidvol.add(m.solidvol[time, l] == m.Y[time, k] / m.vol)
                                else:
                                    new_consolidvol.add(m.solidvol[time, l] == 0.0)
                for jk in list_components:
                    for time in m.huplctime:
                        if jk == k:
                            for l in m.huplc_absorbing.value:
                                strsolidspec = "\'" + str(k) + "\'" + ', ' + "\'" + str(
                                    l) + "\'"  # pair of absorbing solid and liquid
                                if l in m.solid_spec_arg2 and strsolidspec in str(
                                        m.solid_spec.keys()):  #check whether pair of solid and liquid in keys and whether liquid in huplcabs species
                                    valZ = value(m.Z[time, k]) / value(m.vol)
                                    if valZ <= 0:
                                        new_consolidvol.add(m.solidvol[time, l] == 0.0)
                                    else:
                                        new_consolidvol.add(m.solidvol[time, l] == m.Z[time, k] / m.vol)
                                else:
                                    new_consolidvol.add(m.solidvol[time, l] == 0.0)

            m.Dhat_bar = Var(m.huplcmeas_times,
                             m.huplc_absorbing)

            m.Dhat_bar_constraint = Constraint(m.huplcmeas_times,
                                               m.huplc_absorbing,
                                               rule=rule_Dhat_bar)

            for t in m.huplcmeas_times:
                list_huplcabs = [k for k in m.huplc_absorbing.value]
                for k in list_huplcabs:
                    third_term += (m.Dhat[t, k] - m.Dhat_bar[t, k]) ** 2 / sigma_sq['device-huplc']

        return third_term

    def _define_reduce_hess_order(self):
        """
        This sets up the suffixes of the reduced hessian
        """
        self.model.red_hessian = Suffix(direction=Suffix.IMPORT_EXPORT)
        count_vars = 1

        if self._spectra_given:
            if self.model_variance:
                count_vars = self._set_up_reduced_hessian(self.model, self._meas_times, self.component_set, self.component_var, count_vars)
                count_vars = self._set_up_reduced_hessian(self.model, self._meas_lambdas, self.component_set, 'S', count_vars)
                
        for v in self.model.P.values():
            if v.is_fixed():
                print(v, end='\t')
                print("is fixed")
                continue
            self._idx_to_variable[count_vars] = v
            self.model.red_hessian[v] = count_vars
            count_vars += 1
            
        if hasattr(self.model, 'Pinit'):
            for k, v in self.model.Pinit.items():
                v = self.model.init_conditions[k]
                self._idx_to_variable[count_vars] = v
                self.model.red_hessian[v] = count_vars
                count_vars += 1
               
        return None
        
    def _confidence_interval_display(self, variances):
        """
        Function to display calculated confidence intervals
        """
        number_of_stds = st.norm.ppf(1-(1-self.confidence_interval)/2)
        #print(f'STDS: {number_of_stds}')
        
        print('\nConfidence intervals:')
        i = 0
        for k, p in self.model.P.items():
            if p.is_fixed():
                continue
            print('{} ({},{})'.format(k, 
                                      p.value - number_of_stds*(variances[i]**0.5),
                                      p.value + number_of_stds*(variances[i]** 0.5))
                  )
            i += 1
        if hasattr(self.model, 'Pinit'): 
            for k in self.model.Pinit.keys():
                self.model.Pinit[k] = self.model.init_conditions[k].value
                print('{} ({},{})'.format(k, 
                                          self.model.Pinit[k].value - number_of_stds*(variances[i]** 0.5),
                                          self.model.Pinit[k].value + number_of_stds*(variances[i]** 0.5)
                                          )
                      )
                i += 1
        
        return None

    def _compute_covariance(self, hessian, variances):
        """Computes the covariance for post calculation anaylsis
        
        """        
        nt = self._n_allmeas_times
        nw = self._n_meas_lambdas
        nd = nw * nt
        
        if hasattr(self, '_abs_components'):
            isSkipFixed = True
        else:
            isSkipFixed = False

        nparams = self._get_nparams(self.model)
        self._n_params = nparams
        variances_p, covariances_p = self._variances_p_calc(hessian, variances)
        self._confidence_interval_display(variances_p)
        
        return covariances_p

    def _compute_covariance_C_generic(self, hessian, variances, use_model_variance=False):
        """
        Generic covariance function to reduce code
        
        Residuals are not even used...
        """
        if use_model_variance:
            res = self._compute_residuals(self.model)
            nc = self._n_actual
            varmat = np.zeros((nc, nc))
            for c, k in enumerate(self._sublist_components):
                varmat[c, c] = variances[k]
        
        nparams = self._get_nparams(self.model, isSkipFixed=False)
        
        all_H = hessian
        H = all_H[-nparams:, :]

        if not use_model_variance and self._eigredhess2file:
            save_eig_red_hess(H)

        covariance_C = H
        variances_p = np.diag(covariance_C)
        print("Parameter variances: ", variances_p)
        self._confidence_interval_display(variances_p)
       
        return covariance_C

    def _compute_covariance_C(self, hessian, variances):
        """
        Computes the covariance matrix for the paramaters taking in the Hessian
        matrix and the variances for the problem where only C data is provided.
        Outputs the parameter confidence intervals.

        This function is not intended to be used by the users directly

        """
        cov_mat = self._compute_covariance_C_generic(hessian, variances, use_model_variance=True)
        return cov_mat

    def _compute_covariance_no_model_variance(self, hessian, variances):
        """
        Computes the covariance matrix for the paramaters taking in the Hessian
        matrix and the device variance for the problem where model error is ignored.
        Outputs the parameter confidence intervals.

        This function is not intended to be used by the users directly
        """
        cov_mat = self._compute_covariance_C_generic(hessian, variances, use_model_variance=False)
        return cov_mat
    
    def _compute_B_matrix(self, variances, **kwds):
        """Builds B matrix for calculation of covariances

           This method is not intended to be used by users directly

        Args:
            variances (dict): variances

        Returns:
            None
        """
        nt = self._n_meas_times
        time_set = self.model.meas_times
        conc_data_var = 'C'
        nw = self._n_meas_lambdas
        nparams = self._get_nparams(self.model)
       
        if hasattr(self, '_abs_components'):
            n_val = self._nabs_components
            component_set = self._abs_componets
        else:
            n_val = self._n_actual
            component_set = self._sublist_components
    
        variance = variances['device']
        ntheta = n_val * (nw + nt) + nparams
        B_matrix = np.zeros((ntheta, nw * nt))

        for i, t in enumerate(time_set):
            for j, l in enumerate(self.model.meas_lambdas):
                for k, c in enumerate(component_set):
                    r_idx1 = i * n_val + k
                    r_idx2 = j * n_val + k + n_val * nt
                    c_idx = i * nw + j
                    B_matrix[r_idx1, c_idx] = -2 * self.model.S[l, c].value / variance
                    B_matrix[r_idx2, c_idx] = -2 * getattr(self.model, conc_data_var)[t, c].value / variance
                    
        return B_matrix

    def _compute_Vd_matrix(self, variances, **kwds):
        """Builds d covariance matrix

           This method is not intended to be used by users directly

        Args:
            variances (dict): variances

        Returns:
            None
        """
        nt = self._n_meas_times
        nw = self._n_meas_lambdas
        row = []
        col = []
        data = []
        nd = nt * nw
        v_device = variances['device']

        s_array = np.zeros(nw * self.n_val)
        v_array = np.zeros(self.n_val)
        
        for k, c in enumerate(self.component_set):
            v_array[k] = variances[c]

        for j, l in enumerate(self.model.meas_lambdas):
            for k, c in enumerate(self.component_set):
                s_array[j * self.n_val + k] = self.model.S[l, c].value

        for i in range(nt):
            for j in range(nw):
                val = sum(v_array[k] * s_array[j * self.n_val + k] ** 2 for k in range(self.n_val)) + v_device
                row.append(i * nw + j)
                col.append(i * nw + j)
                data.append(val)

                for p in range(nw):
                    if j != p:
                        val = sum(v_array[k] * s_array[j * self.n_val + k] * s_array[p * self.n_val + k] for k in range(self.n_val))
                        row.append(i * nw + j)
                        col.append(i * nw + p)
                        data.append(val)

        Vd_matrix = scipy.sparse.coo_matrix((data, (row, col)), shape=(nd, nd)).tocsr()
        
        return Vd_matrix

    def _calc_new_D(self, subset):
        """Updates the D data for the wavelength selection"""
        
        new_D = pd.DataFrame(np.nan, index=self._meas_times, columns=subset)
        for t in self._meas_times:
            for l in self._meas_lambdas:
                if l in subset:
                    new_D.at[t, l] = self.model.D[t, l]
                    
        return new_D
        

    def run_param_est_with_subset_lambdas(self, builder_clone, end_time, subset, nfe, ncp, sigmas, solver='ipopt', ):
        """ Performs the parameter estimation with a specific subset of wavelengths.
            At the moment, this is performed as a totally new Pyomo model, based on the
            original estimation. Initialization strategies for this will be needed.

                Args:
                    builder_clone (TemplateBuidler): Template builder class of complete model
                                without the data added yet
                    end_time (float): the end time for the data and simulation
                    subset(list): list of selected wavelengths
                    nfe (int): number of finite elements
                    ncp (int): number of collocation points
                    sigmas(dict): dictionary containing the variances, as used in the ParameterEstimator class

                Returns:
                    results (Pyomo model solved): The solved pyomo model

        """
        if not isinstance(subset, (list, dict)):
            raise RuntimeError("subset must be of type list or dict!")

        if isinstance(subset, dict):
            lists1 = sorted(subset.items())
            x1, y1 = zip(*lists1)
            subset = list(x1)
        
        # This is the filter for creating the new data subset
        new_D = self._calc_new_D(subset)
        
        
        print(end_time, new_D)
        # Now that we have a new DataFrame, we need to build the entire problem from this
        # An entire new ParameterEstimation problem should be set up, on the outside of
        # this function and class structure, from the model already developed by the user.
        new_template = construct_model_from_reduced_set(builder_clone, end_time, new_D)
        # need to put in an optional running of the variance estimator for the new
        # parameter estiamtion run, or just use the previous full model run to initialize...

        results, lof = run_param_est(new_template, nfe, ncp, sigmas, solver=solver)

        return results

    def run_lof_analysis(self, builder_before_data, end_time, correlations, lof_full_model, nfe, ncp, sigmas,
                         step_size=0.2, search_range=(0, 1)):
        """ Runs the lack of fit minimization problem used in the Michael's Reaction paper
        from Chen et al. (submitted). To use this function, the full parameter estimation
        problem should be solved first and the correlations for wavelngths from this optimization
        need to be supplied to the function as an option.

                Args:
                    builder_before_data (TemplateBuilder): Template builder class of complete model
                                without the data added yet
                    end_time (int): the end time for the data and simulation

                    correlations (dict): dictionary containing the wavelengths and their correlations
                                to the concentration profiles
                    lof_full_model(int): the value of the lack of fit of the full model (with all wavelengths)

                Returns:
                    *****final model results.

        """
        if not isinstance(step_size, float):
            raise RuntimeError("step_size must be a float between 0 and 1")
        elif step_size >= 1 or step_size <= 0:
            return RuntimeError("step_size must be a float between 0 and 1")

        if not isinstance(search_range, tuple):
            raise RuntimeError("search range must be a tuple")
        elif search_range[0] < 0 or search_range[0] > 1 and not (
                isinstance(search_range, float) or isinstance(search_range, int)):
            raise RuntimeError("search range lower value must be between 0 and 1 and must be type float")
        elif search_range[1] < 0 or search_range[1] > 1 and not (
                isinstance(search_range, float) or isinstance(search_range, int)):
            raise RuntimeError("search range upper value must be between 0 and 1 and must be type float")
        elif search_range[1] <= search_range[0]:
            raise RuntimeError("search_range[1] must be bigger than search_range[0]!")
        # firstly we will run the initial search from at increments of 20 % for the correlations
        # we already have lof(0) so we want 10,30,50,70, 90.
        
        count = 0
        filt = 0.0
        initial_solutions = list()
        initial_solutions.append((0, lof_full_model))
        while filt < search_range[1]:
            filt += step_size
            if filt > search_range[1]:
                break
            elif filt == 1:
                break
            new_subs = wavelength_subset_selection(correlations=correlations, n=filt)
            lists1 = sorted(new_subs.items())
            x1, y1 = zip(*lists1)
            x = list(x1)

            new_D = self._calc_new_D(new_subs)

            # opt_model, nfe, ncp = construct_model_from_reduced_set(builder_before_data, end_time, new_D)
            # Now that we have a new DataFrame, we need to build the entire problem from this
            # An entire new ParameterEstimation problem should be set up, on the outside of
            # this function and class structure, from the model already developed by the user.
            new_template = construct_model_from_reduced_set(builder_before_data, end_time, new_D)
            # need to put in an optional running of the variance estimator for the new
            # parameter estimation run, or just use the previous full model run to initialize...
            results, lof = run_param_est(new_template, nfe, ncp, sigmas)
            initial_solutions.append((filt, lof))

            count += 1

        count = 0
        for x in initial_solutions:
            print("When wavelengths of less than ", x[0], "correlation are removed")
            print("The lack of fit is: ", x[1])

    # =============================================================================
    # --------------------------- DIAGNOSTIC TOOLS ------------------------
    # =============================================================================

    def lack_of_fit(self):
        """ Runs basic post-processing lack of fit analysis

            Args:
                None

            Returns:
                lack of fit (int): percentage lack of fit

        """
        nt = self._n_meas_times
        nw = self._n_meas_lambdas
        sum_e = 0
        sum_d = 0
        C = np.zeros((nt, self.n_val))
        S = np.zeros((nw, self.n_val))
        
        for c_count, c in enumerate(self.component_set):
            for t_count, t in enumerate(self._meas_times):
                C[t_count, c_count] = getattr(self.model, self.component_var)[t, c].value

        for c_count, c in enumerate(self.component_set):
            for l_count, l in enumerate(self._meas_lambdas):
                S[l_count, c_count] = self.model.S[l, c].value
             
        D_model = C.dot(S.T)
        
        for t_count, t in enumerate(self._meas_times):
            for l_count, l in enumerate(self._meas_lambdas):
                sum_e += (D_model[t_count, l_count] - self.model.D[t, l]) ** 2
                sum_d += (self.model.D[t, l]) ** 2
  
        lof = np.sqrt(sum_e/sum_d)*100

        print("The lack of fit is ", lof, " %")
        return lof

    def wavelength_correlation(self):
        """ determines the degree of correlation between the individual wavelengths and
        the and the concentrations.

            Args:
                None

            Returns:
                dictionary of correlations with wavelength

        """
        nt = self._n_meas_times

        cov_d_l = dict()
        for c in self._sublist_components:
            for l in self._meas_lambdas:
                mean_d = (sum(self.model.D[t, l] for t in self._meas_times) / nt)
                mean_c = (sum(self.model.C[t, c].value for t in self._meas_times) / nt)
                cov_d_l[l, c] = 0
                for t in self._meas_times:
                    cov_d_l[l, c] += (self.model.D[t, l] - mean_d) * (self.model.C[t, c].value - mean_c)

                cov_d_l[l, c] = cov_d_l[l, c] / (nt - 1)

        # calculating the standard devs for dl and ck over time
        s_dl = dict()

        for l in self._meas_lambdas:
            s_dl[l] = 0
            mean_d = (sum(self.model.D[t, l] for t in self._meas_times) / nt)
            error = 0
            for t in self._meas_times:
                error += (self.model.D[t, l] - mean_d) ** 2
            s_dl[l] = (error / (nt - 1)) ** 0.5

        s_ck = dict()

        for c in self._sublist_components:
            s_ck[c] = 0
            mean_c = (sum(self.model.C[t, c].value for t in self._meas_times) / nt)
            error = 0
            for t in self._meas_times:
                error += (self.model.C[t, c].value - mean_c) ** 2
            s_ck[c] = (error / (nt - 1)) ** 0.5

        cor_lc = dict()

        for c in self._sublist_components:
            for l in self._meas_lambdas:
                cor_lc[l, c] = cov_d_l[l, c] / (s_ck[c] * s_dl[l])

        cor_l = dict()
        for l in self._meas_lambdas:
            cor_l[l] = max(cor_lc[l, c] for c in self._sublist_components)

        return cor_l

    def lack_of_fit_huplc(self):
        """ Runs basic post-processing lack of fit analysis

            Args:
                None

            Returns:
                lack of fit (int): percentage lack of fit

        """
        nt = self._n_huplcmeas_times
        nc = self._n_huplc
        sum_e = 0
        sum_d = 0
        D_model = np.zeros((nt, nc))
        C = np.zeros((nt, nc))
        
        for t_count, t in enumerate(self._huplcmeas_times):
            for l_count, l in enumerate(self._list_huplcabs):
                
                if hasattr(self.model, 'solidvol'):
                    D_model[t_count, l_count] = (self.model.Z[t, l].value + self.model.solidvol[t, l].value) / (
                        sum(self.model.Z[t, j].value + self.model.solidvol[t, j].value for j in self._list_huplcabs))
                else:
                    D_model[t_count, l_count] = (self.model.Z[t, l].value) / (
                        sum(self.model.Z[t, j].value for j in self._list_huplcabs))

                sum_e += (D_model[t_count, l_count] - self.model.Dhat[t, l].value) ** 2
                sum_d += (self.model.Dhat[t, l].value) ** 2

        lof = np.sqrt((sum_e/sum_d))*100

        print("The lack of fit for the huplc data is ", lof, " %")
        return lof
    
    def g_handling_status_messages(self):
        if self.G_contribution == 'unwanted_G':
            print("\nType of unwanted contributions not set, so assumed that it is time-variant.\n")
            self.G_contribution = 'time_variant_G'
        elif self.G_contribution == 'time_variant_G':
            print("\nTime-variant unwanted contribution is involved.\n")
        elif self.G_contribution == 'time_invariant_G_decompose':
            print("\nTime-invariant unwanted contribution is involved and G can be decomposed.\n")
        elif self.G_contribution == 'time_invariant_G_no_decompose':
            print("\nTime-invariant unwanted contribution is involved but G cannot be decomposed.\n")
        return None
     
    def decompose_G_test(self, St, Z_in):
        """Check whether or not G can be decomposed"""
        
        if St == dict() and Z_in == dict():
            raise RuntimeError('Because time-invariant unwanted contribution is chosen, please provide information of St or Z_in to build omega matrix.')
        
        omega_list = [St[i] for i in St.keys()]
        omega_list += [Z_in[i] for i in Z_in.keys()]
        omega_sub = np.array(omega_list)
        rank = np.linalg.matrix_rank(omega_sub)
        cols = omega_sub.shape[1]
        rko = cols - rank
        
        if rko > 0:
            self.time_invariant_G_decompose = True
            self.G_contribution = 'time_invariant_G_decompose'
        else:
            self.time_invariant_G_no_decompose = True
            self.G_contribution = 'time_invariant_G_no_decompose'
        return None

def wavelength_subset_selection(correlations=None, n=None):
    """ identifies the subset of wavelengths that needs to be chosen, based
    on the minimum correlation value set by the user (or from the automated
    lack of fit minimization procedure)

        Args:
            correlations (dict): dictionary obtained from the wavelength_correlation
                    function, containing every wavelength from the original set and
                    their correlations to the concentration profile.

            n (int): a value between 0 - 1 that slects the minimum amount
                    correlation between the wavelength and the concentrations.

        Returns:
            dictionary of correlations with wavelength

    """
    if not isinstance(correlations, dict):
        raise RuntimeError("correlations must be of type dict()! Use wavelength_correlation function first!")

    # should check whether the dictionary contains all wavelengths or not
    if not isinstance(n, float):
        raise RuntimeError("n must be of type int!")
    elif n > 1 or n < 0:
        raise RuntimeError("n must be a number between 0 and 1")

    subset_dict = dict()
    for l in correlations.keys():
        if correlations[l] >= n:
            subset_dict[l] = correlations[l]
    return subset_dict


# =============================================================================
# ----------- PARAMETER ESTIMATION WITH WAVELENGTH SELECTION ------------------
# =============================================================================

def construct_model_from_reduced_set(builder_clone, end_time, D):
    """ constructs the new pyomo model based on the selected wavelengths.

        Args:
            builder_clone (TemplateBuilder): Template builder class of complete model
                            without the data added yet
            end_time (int): the end time for the data and simulation
            D (dataframe): the new, reduced dataset with only the selected wavelengths.

        Returns:
            opt_mode(TemplateBuilder): new Pyomo model from TemplateBuilder, ready for
                    parameter estimation

    """

    if not isinstance(builder_clone, TemplateBuilder):
        raise RuntimeError('builder_clone needs to be of type TemplateBuilder')

    if not isinstance(D, pd.DataFrame):
        raise RuntimeError('Spectral data format not supported. Try pandas.DataFrame')

    #if not isinstance(end_time, int):
    #    raise RuntimeError('nfe needs to be type int. Number of finite elements must be defined')

    builder_clone._spectral_data = D
    opt_model = builder_clone.create_pyomo_model(0.0, end_time)

    return opt_model


def run_param_est(opt_model, nfe, ncp, sigmas, solver='ipopt'):
    """ Runs the parameter estimator for the selected subset

        Args:
            opt_model (pyomo model): The model that we wish to run the
            nfe (int): number of finite elements
            ncp (int): number of collocation points
            sigmas(dict): dictionary containing the variances, as used in the ParameterEstimator class

        Returns:
            results_pyomo (results of optimization): Parameter Estimation results
            lof (float): lack of fit results

    """

    p_estimator = ParameterEstimator(opt_model)
    p_estimator.apply_discretization('dae.collocation', nfe=nfe, ncp=ncp, scheme='LAGRANGE-RADAU')
    options = dict()

    # These may not always solve, so we need to come up with a decent initialization strategy here
    if solver == 'ipopt':
        results_pyomo = p_estimator.run_opt('ipopt',
                                            tee=False,
                                            solver_opts=options,
                                            variances=sigmas)
    else:
        results_pyomo = p_estimator.run_opt(solver,
                                            tee=False,
                                            solver_opts=options,
                                            variances=sigmas,
                                            covariance=True)
    
    lof = p_estimator.lack_of_fit()

    return results_pyomo, lof


"""
Primary means of using KIPET is through the KipetModel class

@author: kevin 2020
"""
# Standard library imports
import collections
import os
import pathlib
import sys

# Third party imports
import pandas as pd

# Kipet library imports
import kipet.core_methods.data_tools as data_tools
from kipet.core_methods.MEE import MultipleExperimentsEstimator
from kipet.nsd_funs.NSD_KIPET import NSD
from kipet.top_level.reaction_model import (
    ReactionModel,
    _set_directory,
    )
from kipet.top_level.settings import (
    Settings, 
    )
from kipet.top_level.unit_base import UnitBase

# from kipet.top_level.element_blocks import (
#     AlgebraicBlock,
#     ComponentBlock,
#     ConstantBlock, 
#     ParameterBlock, 
#     StateBlock,
#     )
from kipet.core_methods.TemplateBuilder import TemplateBuilder

class KipetModel():
    
    """Primary KIPET object that holds ReactionModel instances"""
    
    def __init__(self):
        
        self.models = {}
        self.settings = Settings()
        self.results = {}
        self.global_parameters = None
        self.method = 'mee'
        
        self.use_individual_settings = False
        # ub = UnitBase()
        self.ub = UnitBase()
        self.reset_base_units()    
    
        # self.components = ComponentBlock()   
        # self.parameters = ParameterBlock()
        # self.constants = ConstantBlock()
        # self.algebraics = AlgebraicBlock()
        # self.states = StateBlock()
    
        
    def __str__(self):
        
        block_str = "KipetModel\n\n"
        
        for name, model in self.models.items():
            block_str += f'{name}\tDatasets: {len(model.datasets)}\n'
        
        return block_str

    def __repr__(self):
        return self.__str__()

    def __iter__(self):
        for model, contents in self.models.items():
            yield data
            
    def __getitem__(self, value):
        return self.models[value]
            
    def __len__(self):
        return len(self.models)

    def reset_base_units(self):
        
        self.ub.TIME_BASE = self.settings.units.time
        self.ub.VOLUME_BASE = self.settings.units.volume

    def add_model_list(self, model_list):
        """Handles lists of parameters or single parameters added to the model
        """
        for model in model_list:
            self.add_model(model)         
        
        return None
    
    def add_model(self, model):
        """Redundant"""
        self.add_reaction(model)  
    #     if isinstance(model, ReactionModel):
    #         self.models[model.name] = model
    #     else:
    #         raise ValueError('KipetModel can only add ReactionModel instances.')
            
    #     return None
    
    def remove_model(self, model):
        
        if isinstance(model, str):
            if model in self.models:
                self.models.pop(model)
        elif isinstance(model, ReactionModel):
            self.models.pop(model.name)
        else:
            print('KipetModel does not have specified model')
        return None    
    
    def new_reaction(self, name, model=None, ignore=None):
        
        """New reactions can be added to KIPET using this function
        
        Args:
            name (str): The name of the model/experiment used in all references
                made to it in KIPET, especially in python dicts
            model_to_clone (ReactionModel): You can copy an existing ReactionModel by
                adding it here
            items_not_copied (list): This is a list of strings for the various
                attributes in the ReactionModel that should not be copied
                example: ['datasets'] ==> does not copy the dataset into the 
                new model
                
        Returns:
            self.models[name] (ReactionModel): This is the newly created
                ReactionModel instance
                
        """
        ignore = ignore if ignore is not None else []
        
        if model is None:
        
            self.models[name] = ReactionModel(name=name, unit_base=self.ub)
#            self.models[name].settings.general.data_directory = self.settings.general.data_directory
            
            assign_list = ['components', 'parameters', 'constants', 'algebraics',
                             'states', 'ub', 'c']
        
            for item in assign_list:
                if item not in ignore and hasattr(self, item):
                    setattr(self.models[name], item, getattr(self, item))
        
        else:
            if isinstance(model, ReactionModel):
                kwargs = {}
                kwargs['name'] = name
                self.models[name] = ReactionModel(name=name, unit_base=self.ub)
                
                assign_list = ['components', 'parameters', 'constants', 'algebraics',
                             'states', 'ub', 'settings', 'c', 'odes_dict']
        
              #  copy_list = []#'c', 'odes_dict']# 'set_up_model']
                for item in assign_list:
                    if item not in ignore and hasattr(model, item):
                        setattr(self.models[name], item, getattr(model, item))
                
                # import copy
                # for item in copy_list:
                #     if item not in ignore and hasattr(model, item):
                #         setattr(self.models[name], item, copy.copy(getattr(model, item)))
                
            else:
                raise ValueError('KipetModel can only add ReactionModel instances.')
        
        return self.models[name]
    
    def add_reaction(self, kipet_model_instance):
    
        if isinstance(kipet_model_instance, ReactionModel):
            kipet_model_instance.unit_base = self.ub
            self.models[kipet_model_instance.name] = kipet_model_instance
            #self.models[kipet_model_instance.name].settings.general.data_directory = self.settings.general.data_directory
        else:
            raise ValueError('KipetModel can only add ReactionModel instances.')

        return None
    
    # def constant(self, *args, **kwargs):
        
    #     kwargs['unit_base'] = self.ub
    #     self.constants.add_element(*args, **kwargs)
    #     return None
        
    # def algebraic(self, *args, **kwargs):
        
    #     kwargs['unit_base'] = self.ub
    #     self.algebraics.add_element(*args, **kwargs)
    #     # if 'step' in kwargs and kwargs['step'] is not None:
    #     #     self.add_step(f's_{args[0]}', time=15, switch='off')
    
    # # def add_step(self, name, *args, **kwargs):
        
    # #     self._has_step_or_dosing = True
    # #     if not hasattr(self, '_step_list'):
    # #         self._step_list = {}
            
    # #     if name not in self._step_list:
    # #         self._step_list[name] = [kwargs]
    # #     else:
    # #         self._step_list[name].append(kwargs)
    
    # def state(self, *args, **kwargs):
    #     """Add the components to the Kipet instance
        
    #     Args:
    #         components (list): list of Component instances
            
    #     Returns:
    #         None
            
    #     """
    #     kwargs['unit_base'] = self.ub
    #     self.states.add_element(*args, **kwargs)
    #     return None
    
    # def component(self, *args, **kwargs):
    #     """Add the components to the Kipet instance
        
    #     Args:
    #         components (list): list of Component instances
            
    #     Returns:
    #         None
            
    #     """
    #     kwargs['unit_base'] = self.ub
    #     self.components.add_element(*args, **kwargs)
    #     return None
    
    # def parameter(self, *args, **kwargs):
    #     """Add the parameters to the Kipet instance
        
    #     Args:
    #         parameters (list): list of Parameter instances
            
    #         factor (float): defaults to 1, the scalar multiple of the parameters
    #         for simulation purposes
            
    #     Returns:
    #         None
            
    #     """
    #     kwargs['unit_base'] = self.ub
    #     self.parameters.add_element(*args, **kwargs)
    #     return None
    
    # def get_model_vars(self):
        
    #     self.create_pyomo_model_vars()
    #     return self.c
    
    # # def load_vars(self):
        
    # #     for var in self.c.get_var_list:
    # #         globals()[var] = getattr(self.c, var)
    
    # def create_pyomo_model_vars(self, *args, **kwargs):
        
    #     self.builder = TemplateBuilder()
    #     setattr(self.builder, 'early_return', True)
            
    #     if len(self.states) > 0:
    #         self.builder.add_model_element(self.states)
        
    #     if len(self.algebraics) > 0:
    #         self.builder.add_model_element(self.algebraics)
        
    #     if len(self.components) > 0:
    #         self.builder.add_model_element(self.components)
            
    #     if len(self.parameters) > 0:
    #         self.builder.add_model_element(self.parameters)
        
    #     if len(self.constants) > 0:
    #         self.builder.add_model_element(self.constants)
        
    #     if hasattr(self, 'c'):
    #         setattr(self.builder, 'c_mod', self.c)
        
    #     self.model = self.builder.create_pyomo_model(0, 1)
    #     self.c = self.builder.c_mod
    #     #self.set_up_model = self.model
    #     self.model = None
            
    #     return None
    
            
    def run_opt(self, *args, **kwargs):
        """Solve a single model or solve multiple models using the MEE
        """
        method = kwargs.get('method', 'mee')
        
        if len(self.models) > 1:
            if method == 'mee':
                self._calculate_parameters()
                self._create_multiple_experiments_estimator(*args, **kwargs)
                self.run_full_model()
            elif method == 'nsd':
                self._calculate_parameters()
                self.mee_nsd(strategy='ipopt')
            else:
                raise ValueError('Not a valid method for optimization')
            
        else:
            reaction_model = self.models[list(self.models.keys())[0]]
            results = reaction_model.run_opt()
            self.results[reaction_model.name] = results
            
        return None
    
    def _create_multiple_experiments_estimator(self, *args, **kwargs):
        """A quick wrapper for MEE without big changes
        
        """
        self.mee = MultipleExperimentsEstimator(self.models)
        self.mee.confidence_interval = self.settings.general.confidence
        
        if 'spectral' in self.data_types:
            self.settings.general.spectra_problem = True
        else:
            self.settings.general.spectra_problem = False    
        
        self.mee.spectra_problem = self.settings.general.spectra_problem
        
    def _calculate_parameters(self):
        """Uses the ReactionModel framework to calculate parameters instead of 
        repeating this in the MEE
        
        """
        for name, model in self.models.items():
            if not model.optimized:
                # Do this for each model separately...
                # for dataset in model.datasets:
                #     if self.settings.general.freq_wavelength_subset is not None:
                #         if model.datasets[dataset.name].category == 'spectral':
                #             freq = self.settings.general.freq_wavelength_subset
                #             model.datasets[dataset.name].data = decrease_wavelengths(dataset.data, freq)
                
                # if not self.use_individual_settings:
                    
                #     model.settings.update(**self.settings)
                
                model.run_opt()
            else:
                print(f'Model {name} has already been optimized')
                
        return None
    
    def run_full_model(self):
        
        # self.settings.parameter_estimator.solver_opts = self.settings.solver
        consolidated_settings = self.settings.general
        consolidated_settings.solver_opts = self.settings.solver
        
        results = self.mee.solve_consolidated_model(self.global_parameters,
                                                    **consolidated_settings)
    
        self.results = results
        for key, results_obj in self.results.items():
#            results_obj.file_dir = self.settings.general.charts_directory
            self.models[key].results = results[key]
            
        return results
    
    def mee_nsd(self, strategy='ipopt'):
        """Performs the NSD on the multiple datasets
        
        Args:
            strategy (str): Method used to control the outer problem
                ipopt, trust-region, newton-step
                
        Returns:
            results
        
        """        
        kwargs = {'kipet': True,
                  'objective_multiplier': 1
                  }
        
        if self.global_parameters is not None:
            global_parameters = self.global_parameters
        else:
            global_parameters = self.all_params
        
        self.nsd = NSD(self.models,
                       strategy=strategy,
                       global_parameters=global_parameters, 
                       kwargs=kwargs)
        
        print(self.nsd.d_init)
        
        results = self.nsd.run_opt()
        
        self.results = results
        for key, results_obj in self.results.items():
            results_obj.file_dir = self.settings.general.charts_directory
            
        return results
    
    def write_data_file(self, filename, data, directory=None, filetype='csv'):
        """Method to write data to a file using KipetModel
        
        Args:
            filename (str): the name of the file (plus relative directory)
            
            data (pandas DataFrame): The data to be written to the file
            
            directory (str): absolute directory to use instead
            
            filetype (str): the filetype to be used (in case not in file name)
        
        Returns:
            None
        
        """
        _filename = filename
        # if directory is None:
        #     _filename = _set_directory(self, filename)
        # else:
        #     _filename = pathlib.Path(directory).joinpath(filename)
        # if not _filename.parent.is_dir():
        #     _filename.parent.mkdir(exist_ok=True)
        calling_file_name = os.path.dirname(os.path.realpath(sys.argv[0]))
        _filename = pathlib.Path(calling_file_name).joinpath(_filename)
        data_tools.write_file(_filename, data)
        
        return None
        
    def read_data_file(self, filename, directory=None):
        """Method to read data file using KipetModel
        
        Args:
            filename (str): the name of the file (plus relative directory)
            
            data (pandas DataFrame): The data to be written to the file
            
            directory (str): absolute directory to use instead
            
        Returns:
            read_data (pandas DataFrame): The data read from the file
        
        """
        _filename = filename
        # if directory is None:
        #     _filename = _set_directory(self, filename)
        # else:
        #     _filename = pathlib.Path(directory).joinpath(filename)
        
        calling_file_name = os.path.dirname(os.path.realpath(sys.argv[0]))
        _filename = pathlib.Path(calling_file_name).joinpath(_filename)
        read_data = data_tools.read_file(_filename)
        
        return read_data
    
    def plot(self, *args, **kwargs):
        """Plots the modeled concentration profiles"""
        
        for reaction, result in self.results.items():
            description={'title': f'Experiment: {reaction}',
                                  'xaxis': 'Time [s]',
                                  'yaxis': 'Concentration [mol/L]'}
        
            result.plot('Z', description=description)
            
        return None
    
    @staticmethod
    def add_noise_to_data(data, noise):
        """Wrapper for adding noise to data after data has been added to
        the specific ReactionModel
        
        """
        return data_tools.add_noise_to_signal(data, noise)   
    
    # def delete_file(self, filename, directory=None):
    #     """Method to remove files from the directory"""
    #     remove_file(self.settings.general.data_directory, 'ipopt_opt')
    
    @property
    def all_params(self):
        set_of_all_model_params = set()
        for name, model in self.models.items():
            set_of_all_model_params = set_of_all_model_params.union(model.parameters.names)
            
        return set_of_all_model_params
   
    @property
    def model_list(self):
        return [model.name for model in self.models.values()]
    
    @property
    def data_types(self):
        data_types = set()
        for name, kipet_model in self.models.items():
            for dataset_name, dataset in kipet_model.datasets.datasets.items():
                data_types.add(dataset.category)
                
        return data_types
    
    @property
    def show_parameters(self):
        
        df_param = pd.DataFrame(data=None, index=self.all_params, columns=self.models.keys())
            
        for reaction, model in self.models.items():
            for param in model.parameters.names:
                df_param.loc[param, reaction] = model.results.P[param]
            
        return df_param
           
    @property
    def get_p_models(self):
        
        return [model.p_model for model in self.models.values()]
    
    
    # Testing placing components into the KM
    
    
    
    
    
    
    


"""

This is a new NSD module for use with IPOPT

Goal: To be clean as possible with the implementation and to use as many of
      the existing code as possible - it is all in the reduced_hessian already.
"""
# Standard library imports
import copy

# Third party imports
import ipopt
import numpy as np
import pandas as pd
import plotly.graph_objects as go
from plotly.offline import download_plotlyjs, init_notebook_mode, plot, iplot
from pyomo.environ import (
    Objective,
    SolverFactory,
    Suffix,
    )
from scipy.optimize import (
    Bounds,
    minimize,
    )

from scipy.sparse import coo_matrix

# KIPET library imports
from kipet.common.parameter_handling import (
    check_initial_parameter_values,
    set_scaled_parameter_bounds,
    )
from kipet.common.parameter_ranking import (
    parameter_ratios,
    rank_parameters,
    )
# from kipet.common.reduced_hessian import (
#     add_global_constraints,
#     calculate_reduced_hessian,
#     optimize_model,
#     calculate_duals,
#     )
from kipet.post_model_build.scaling import (
    remove_scaling,
    scale_parameters,
    update_expression,
    )
from kipet.common.objectives import (
    conc_objective,
    comp_objective,
    )
from kipet.core_methods.ParameterEstimator import ParameterEstimator
from kipet.common.ReducedHessian import ReducedHessian
        
DEBUG = False

class NSD():
    
    def __init__(self, model_list, init=None, kwargs=None):
        
        kwargs = kwargs if kwargs is not None else {}
        parameter_name = kwargs.get('parameter_name', 'P')
        self.objective_multiplier = kwargs.get('objective_multiplier', 1)
        self.scaled = kwargs.get('scaled', False)
        self.isKipetModel = kwargs.get('kipet', True)
        
        self.reduced_hessian_kwargs = {}
        
        if self.isKipetModel:
            self.kipet_models = model_list
            self.model_list = [r.model for r in model_list]
        
        else:
            self.model_list = model_list
        
        #self.model_list = model_list
        
        model_one = self.model_list[0]
        
        print(model_one)
        # if not self.scaled:    
        #     self.d_init = {p: [0.5*(v.lb + v.ub), v.lb, v.ub] for p, v in getattr(model_one, parameter_name).items()}
        # else:
        #     
        self.d_init = {p: [1.0*v.value, v.lb, v.ub] for p, v in getattr(model_one, parameter_name).items()}
            
        #self.d_init['k'] = [1.1, 0.1, 10]
        
        if init is not None:
            for k, v in init.items():
                self.d_init[k][0] = v
        
        print(self.d_init)
        
        self.method = 'trust-constr'
        
        all_parameters = []
        for model in self.model_list:
            for param in model.P.keys():
                if param not in all_parameters:
                    all_parameters.append(param)
        
        self.parameter_names = all_parameters
        
    def __str__(self):
        
        return 'Nested Schur Decomposition Object'
    
    def set_initial_value(self, init):
        """Add custom initial values for the parameters
        
        Args:
            init (dict): keys are parameters, values are floats
            
        Returns:
            None
            
        """
        for k, v in init.items():
            if k in self.d_init:
                self.d_init[k][0] = v
    
        return None
    
    def _rule_objective(self, model):
        """This function defines the objective function for the given model
        
        Args:
            model (pyomo.core.base.PyomoModel.ConcreteModel): This is the pyomo
            model instance for the estimability problem.
                
        Returns:
            obj (pyomo.environ.Objective): This returns the objective function
            for the estimability optimization.
        
        """
        obj = 0
        obj += 0.5*conc_objective(model)*self.objective_multiplier
        obj += 0.5*comp_objective(model)*self.objective_multiplier
    
        return Objective(expr=obj)
    
    def _model_preparation(self, scaled=True, use_duals=True):
        """Helper function that should prepare the models when called from the
        main function. Includes the experimental data, sets the objectives,
        simulates to warm start the models if no data is provided, sets up the
        reduced hessian model with "fake data", and discretizes all models

        """
        for model in self.model_list:
            
            if not hasattr(model, 'objective'):
                model.objective = self._rule_objective(model)
            
            # The model needs to be discretized
            model_pe = ParameterEstimator(model)
            model_pe.apply_discretization('dae.collocation',
                                          ncp=3,  #self.ncp,
                                          nfe=50, #self.nfe,
                                          scheme='LAGRANGE-RADAU')
            
            # Here is where the parameters are scaled
            #if scaled:
            #    scale_parameters(model)
            #    set_scaled_parameter_bounds(model, rho=10) #self.rho)
            
            #else:
            #check_initial_parameter_values(model)
            
            if use_duals:
                model.dual = Suffix(direction=Suffix.IMPORT_EXPORT)
                model.ipopt_zL_out = Suffix(direction=Suffix.IMPORT)
                model.ipopt_zU_out = Suffix(direction=Suffix.IMPORT)
                model.ipopt_zL_in = Suffix(direction=Suffix.EXPORT)
                model.ipopt_zU_in = Suffix(direction=Suffix.EXPORT)
                # model.red_hessian = Suffix(direction=Suffix.EXPORT)
                # model.dof_v = Suffix(direction=Suffix.EXPORT)
                # model.rh_name = Suffix(direction=Suffix.IMPORT)
                
                # count_vars = 1
                # for k, v in model.P.items():
                #     model.dof_v[k] = count_vars
                #     count_vars += 1
                
                # model.npdp = Suffix(direction=Suffix.EXPORT)
            
        return None
    
    # def _model_preparation(self):
    #     """Helper function that should prepare the models when called from the
    #     main function. Includes the experimental data, sets the objectives,
    #     simulates to warm start the models if no data is provided, sets up the
    #     reduced hessian model with "fake data", and discretizes all models

    #     """
    #     if not hasattr(self.model, 'objective'):
    #         self.model.objective = self._rule_objective(self.model)
        
    #     # The model needs to be discretized
    #     model_pe = ParameterEstimator(self.model)
    #     model_pe.apply_discretization('dae.collocation',
    #                                   ncp=self.ncp,
    #                                   nfe=self.nfe,
    #                                   scheme='LAGRANGE-RADAU')
        
    #     # Here is where the parameters are scaled
    #     if self.scaled:
    #         scale_parameters(self.model)
    #         set_scaled_parameter_bounds(self.model, rho=self.rho)
        
    #     else:
    #         check_initial_parameter_values(self.model)
        
    #     if self.use_duals:
    #         self.model.dual = Suffix(direction=Suffix.IMPORT_EXPORT)
        
    #     return None
    
    def _generate_bounds_object(self):
        """Creates the Bounds object needed by SciPy for minimization
        
        Returns:
            bounds (scipy Bounds object): returns the parameter bounds for the
                trust-region method
        
        """
        lower_bounds = []
        upper_bounds = []
        
        for k, v in self.d_init.items():
            lower_bounds.append(v[1])
            upper_bounds.append(v[2])
        
        lower_bounds = np.array(lower_bounds, dtype=float)
        upper_bounds = np.array(upper_bounds, dtype=float) 
        bounds = Bounds(lower_bounds, upper_bounds, True)
        return bounds
    
    @staticmethod
    def objective_function(x, scenarios, parameter_names):
        """Inner problem calculation for the NSD
        
        Args:
            x (np.array): array of parameter values
            
            scenarios (list): list of reaction models
            
            parameter_names (list): list of global parameters
            
        Returns:
            
            objective_value (float): sum of sub-problem objectives
        """
        if DEBUG:
            stuck = '*'*50
            print(stuck)
            print('\nCalculating objective')
            
        
        for i, p in enumerate(parameter_names):
            print(f'{p} = {x[i]:0.12f}')
        
        objective_value = 0
        for i, model in enumerate(scenarios):
            
            rh = ReducedHessian(model, file_number=i)
            rh.parameter_set = parameter_names
            rh.optimize_model(d=x)
            objective_value += model.objective.expr()
        
        if DEBUG:
            print(f'Obj: {objective_value}')
            print(stuck)
        return objective_value
    
    @staticmethod
    def calculate_m(x, scenarios, parameter_names):
        """Calculate the vector of duals for the NSD
        
        Args:
            x (np.array): array of parameter values
            
            scenarios (list): list of reaction models
            
            parameter_names (list): list of global parameters
            
        Returns:
            
            m (np.array): vector of duals
            
        """
        if DEBUG:
            stuck = '*'*50
            print(stuck)
            print('\nCalculating m')
        
        m = pd.DataFrame(np.zeros((len(parameter_names), 1)), index=parameter_names, columns=['dual'])
        
        kwargs = {
            'param_con_method': 'global',
            'kkt_method': 'k_aug',
            'set_param_bounds': False,
            'param_set_name': 'parameter_names',
            }
        
        for i, model_opt in enumerate(scenarios):
            
            rh = ReducedHessian(model_opt, file_number=i, **kwargs)
            rh.parameter_set = parameter_names
            
            if not hasattr(model_opt, 'd'):
                rh.optimize_model(d=x)
            
            duals = rh.calculate_duals() 
            for param in m.index:
                if param in duals.keys():
                    m.loc[param] = m.loc[param] + duals[param]

            rh.delete_sol_files()

        m = m.values.flatten()
        
        if DEBUG:
            print(f'm: {m}')
            print(stuck)
        
        return m
    
    @staticmethod
    def calculate_M(x, scenarios, parameter_names):
        """Calculate the sum of reduced Hessians for the NSD
        
        Args:
            x (np.array): array of parameter values
            
            scenarios (list): list of reaction models
            
            parameter_names (list): list of global parameters
            
        Returns:
            
            M (np.array): sum of reduced Hessians
        """
        if DEBUG:
            stuck = '*'*50
            print(stuck)
            print('\nCalculating M')
            
        M_size = len(parameter_names)
        M = pd.DataFrame(np.zeros((M_size, M_size)), index=parameter_names, columns=parameter_names)
            
        for i, model in enumerate(scenarios):
            
            rh = ReducedHessian(model, file_number=i)
            rh.parameter_set = parameter_names
            reduced_hessian = rh.calculate_reduced_hessian()
            
            M = M.add(reduced_hessian).combine_first(M)
            M = M[parameter_names]
            M = M.reindex(parameter_names)
        
        M = M.values# + np.eye(M_size)*0.1
        
        if DEBUG:
            print(f'M: {M}')
            print(f'Det:  {np.linalg.det(M):0.4f}')
            print(f'Rank: {np.linalg.matrix_rank(M)}')
            print(f'EigVals: {np.linalg.eigh(M)[0]}')
            print(stuck)
        
        return M
    
    @staticmethod
    def calculate_grad(x, scenarios, parameter_names):
        """Calculate the average of the gradients for the NSD
        
        Args:
            x (np.array): array of parameter values
            
            scenarios (list): list of reaction models
            
            parameter_names (list): list of global parameters
            
        Returns:
            
            M (np.array): sum of reduced Hessians
        """
        return np.zeros((len(x), 1))
       
    def ipopt_method(self, scaled=False, callback=None, options=None, **kwargs):
        """ Minimization of scalar function of one or more variables with
            constraints
    
        Args:
            m : PyomoNLP Model or equivalent
    
            callback  : callable
                Called after each iteration.
    
                    ``callback(xk, state) -> bool``
                
                where ``xk`` is the current parameter vector. and ``state`` is
                an optimization result object. If callback returns True, the algorithm
                execution is terminated.
    
            options : IPOPT options
        
        Returns:
            result : Optimization result
        
        """
        # Prepare the models for NSD
        if self.isKipetModel:
            self._model_preparation(scaled=self.scaled, use_duals=True)
        
        for model in self.model_list:
            for param, model_param in model.P.items():
                model_param.value = self.d_init[param][0]
        
        # Set up the initial parameter values
        d_init = self.d_init
        d_vals =  [d[0] for k, d in self.d_init.items()]
        print(f'd_vals: {d_vals}')
        if scaled:
            # d_init_unscaled = {}
            d_vals = [1 for p in d_vals]
    
        kwargs = {
                'scenarios': self.model_list,
                'parameter_names': self.parameter_names,
                'parameter_number': len(d_vals)
                 }
    
        problem_object = Optproblem(objective=self.objective_function,
                                    hessian=self.calculate_M,
                                    gradient=self.calculate_m,
                                    jacobian=self.calculate_grad,
                                    kwargs=kwargs,
                                    callback=callback)
        
        bounds = self._generate_bounds_object()
        print(bounds)
        
        nlp = ipopt.problem(n = len(d_vals),
                            m = 0,
                            problem_obj = problem_object,
                            lb = bounds.lb,
                            ub = bounds.ub,
                            cl = [],
                            cu = [],
                            )
    
        options = {'tol': 1e-8, 
                 #  'bound_relax_factor': 1.0e-8, 
                   'max_iter': 100,
                   'print_user_options': 'yes', 
                   'nlp_scaling_method': 'none',
                   #'corrector_type': 'primal-dual',
                   #'alpha_for_y': 'full',
               #    'accept_every_trial_step': 'yes',
                  # 'linear_solver': 'ma57'
                   }

        if options: 
            for key, value in options.items():
                nlp.addOption(key, value)
        
        
        x, results = nlp.solve(d_vals)
        
        # Prepare parameter results
        # print(d_init_unscaled)
        # if scaled:
        #     s_factor = {k: d_init_unscaled[k] for k in self.d_init.keys()}
        # else:
        s_factor = {k: 1 for k in self.d_init.keys()}
        
        self.parameters_opt = {k: results['x'][i]*s_factor[k] for i, k in enumerate(self.d_init.keys())}
    
        return results
    
    def trust_region(self, debug=False, scaled=False):
        """This is the outer problem controlled by a trust region solver 
        running on scipy. This is the only method that the user needs to 
        call after the NSD instance is initialized.
        
        Returns:
            results (scipy.optimize.optimize.OptimizeResult): The results from the 
                trust region optimation (outer problem)
                
            opt_dict (dict): Information obtained in each iteration (use for
                debugging)
                
        """
        # Prepare the models for NSD
        if self.isKipetModel:
            self._model_preparation(scaled=scaled, use_duals=True)
        
        # Set up the initial parameter values
        d_init = self.d_init
        d_vals =  [d[0] for k, d in self.d_init.items()]
        if scaled:
            d_init_unscaled = d_vals
            d_vals = [1 for p in d_vals]

        # Record the parameter values in each iteration
        self.d_iter = []
        def callback(x, *args):
            self.d_iter.append(x)
    
        # Start TR Routine
        if self.method in ['trust-exact', 'trust-constr']:

            tr_options={
                #'xtol': 1e-6,
                }
            
            results = minimize(self.objective_function, 
                                d_vals,
                                args=(self.model_list, self.parameter_names), 
                                method=self.method,
                                jac=self.calculate_m,
                                hess=self.calculate_M,
                                callback=callback,
                                bounds=self._generate_bounds_object(),
                                options=tr_options,
                            )
            
            # Prepare parameter results
            if scaled:
                s_factor = {k: d_init_unscaled[k] for k in self.d_init.keys()}
            else:
                s_factor = {k: 1 for k in self.d_init.keys()}
            
            self.parameters_opt = {k: results.x[i]*s_factor[k] for i, k in enumerate(self.d_init.keys())}
       
        return results
    
    def run_simple_newton_step(self, debug=False, scaled=False, alpha=1, iterations=15, opt_tol=1e-8):
        
        # Prepare the models for NSD
        if self.isKipetModel:
            self._model_preparation(scaled=scaled, use_duals=True)
        
        for model in self.model_list:
            for param, model_param in model.P.items():
                model_param.value = self.d_init[param][0]
        
        # options = dict(calc_method='global',
        #                 method='k_aug', 
        #                 scaled=scaled,
        #                 use_duals=use_duals,
        #                 set_up_constraints=ADD_CONSTRAINTS,
        #                 )
    
        d_init = self.d_init
        d_vals =  [d[0] for k, d in self.d_init.items()]
        if scaled:
            d_init_unscaled = d_vals
            d_vals = [1 for p in d_vals]
    
        for i in range(iterations):
            
            obj_val = self.objective_function(
                                    d_vals,
                                    self.model_list, 
                                    self.parameter_names,
                                    )
           
            # Get the M matrices to determine search direction
            M = self.calculate_M(d_vals, self.model_list, self.parameter_names)
            m = self.calculate_m(d_vals, self.model_list, self.parameter_names)
            
            # Calculate the search direction
            d = np.linalg.inv(M) @ -(m)
            
            print(f'd: {d}')
            
            # Update model parameters
            for model in self.model_list:
                for j, param in enumerate(self.parameter_names):
                    model.P[param].set_value(d[j]*alpha + model.P[param].value)
                    d_vals = d*alpha + d_vals
                
            model.P .display()
            if max(abs(d)) <= opt_tol:
                print('Tolerance reached')
                break
        
        # Update ReactionModel objects with the final parameter values
        if self.isKipetModel:
            for m, reaction in enumerate(self.kipet_models):
                for k, v in reaction.model.P.items():
                    if scaled:
                        reaction.model.P[k].set_value(self.model_list[m].K[k].value*self.model_list[m].P[k].value)
                    else:
                        reaction.model.P[k].set_value(self.model_list[m].P[k].value)
            
        # Prepare parameter results
        if scaled:
            s_factor = {k: d_init_unscaled[k] for k in self.d_init.keys()}
        else:
            s_factor = {k: 1 for k in self.d_init.keys()}
            
        self.parameters_opt = {k: d_vals[i]*s_factor[k] for i, k in enumerate(self.d_init.keys())}
        # Plot the results of the fitting
        self.plot_results()
        
        return None
    
    def plot_paths(self, filename=''):
        """Plot the parameter paths through parameter space during the NSD
        algorithm. For diagnostic purposes.
        
        """
        x_data = list(range(1, len(self.d_iter) + 1))
        y_data = np.r_[self.d_iter]
        
        fig = go.Figure()    
        
        for i, params in enumerate(self.d_init.keys()):
            
            fig.add_trace(
                go.Scatter(x = x_data,
                           y = y_data[:, i],
                           name = params,
                      # line=dict(color=colors[i], width=4),
                   )
                )
        
        fig.update_layout(
            title='Parameter Paths in NSD',
            xaxis_title='Iterations',
            yaxis_title='Parameter Values',
            )
    
        plot(fig)
    
        return None
                
    def plot_results(self):
        
        if self.isKipetModel:
        
            for model in self.kipet_models:
                model.simulate()
                model.results.plot(extra_data={'data': model.datasets['C_data'].data, 'label': 'Meas.', 'mode': 'marker'})        
                #model.results.plot(extra_data={'data': model.datasets['U_data'].data, 'label': 'Meas.'})        
        
        
        return None
            
class Optproblem(object):
    """Optimization problem

    This class defines the optimization problem which is callable from cyipopt.

    """
    def __init__(self, 
                 objective=None, 
                 hessian=None, 
                 jacobian=None, 
                 gradient=None, 
                 kwargs={}, 
                 callback=None):
        
        self.fun = objective
        self.grad = gradient
        self.hess = hessian
        self.jac = jacobian
        self.kwargs = kwargs

    def objective(self, x):
        
        scenarios = self.kwargs.get('scenarios', None)
        parameter_names = self.kwargs.get('parameter_names', None)
        
        return self.fun(x, scenarios, parameter_names)
    
    def gradient(self, x):
        
        scenarios = self.kwargs.get('scenarios', None)
        parameter_names = self.kwargs.get('parameter_names', None)
        
        return self.grad(x, scenarios, parameter_names)

    def constraints(self, x):
        """The problem is unconstrained in the outer problem excluding
        parameters
        
        """
        return np.array([])

    def jacobian(self, x):

        scenarios = self.kwargs.get('scenarios', None)
        parameter_names = self.kwargs.get('parameter_names', None)

        return self.jac(x, scenarios, parameter_names)        

    def hessianstructure(self):
        
        global hs
        nx = self.kwargs['parameter_number']
        hs = coo_matrix(np.tril(np.ones((nx, nx))))
        
        return (hs.col, hs.row)
        
    def hessian(self, x, a, b):
        
        scenarios = self.kwargs.get('scenarios', None)
        parameter_names = self.kwargs.get('parameter_names', None)
        H = self.hess(x, scenarios, parameter_names)
        
        return H[hs.row, hs.col]

    
if __name__ == '__main__':
   
    from kipet.new_examples.Ex_18_NSD import generate_models, generate_models_cstr
    
    # Generate the ReactionModels (at some point KipetModel)
    models = generate_models()

    # Create the NSD object using the ReactionModels list
    kwargs = {'kipet': True,
              'objective_multiplier': 1
              }
    # The objective_multiplier is helpful when the objective is already small
    
    # Choose the method used to optimize the outer problem
    strategy = 'ipopt'
    #strategy = 'newton-step'
    #strategy = 'trust-region'
    
    nsd = NSD(models, kwargs=kwargs)
    
    # Set the initial values
    nsd.set_initial_value({'k1' : 0.6,
                           'k2' : 1.2}
                          )
    
    # nsd.set_initial_value({'Cfa' : 1.500000,
    #                         'rho' : 1.600000,
    #                         'ER' : 1.100000,
    #                         'k' : 1.100000,
    #                         'Tfc' : 1.100000,
    #                         }
    #                       )
    
    print(nsd.d_init)
    
    if strategy == 'ipopt':
        # Runs the IPOPT Method
        results = nsd.ipopt_method(scaled=True)
    
    elif strategy == 'trust-region':
        # Runs the Trust-Region Method
        results = nsd.trust_region(scaled=False)
        # Plot the parameter value paths (Trust-Region only)
        nsd.plot_paths()
        
    elif strategy == 'newton-step':
        # Runs the NSD using simple newton steps
        nsd.run_simple_newton_step(alpha=0.1, iterations=15)  
    
    # Plot the results using ReactionModel format
    nsd.plot_results()



"""
Module to hold the method developed by Chen et al. 2016
"""
import numpy as np
import scipy
from scipy.optimize import least_squares

from pyomo.environ import (
    Objective,
    SolverFactory,
    )

def solve_C(var_est_object, solver, **kwds):
     """Solves formulation 23 from Weifengs procedure with ipopt

        This method is not intended to be used by users directly

     Args:

     Returns:
         None

     """
     solver_opts = kwds.pop('solver_opts', dict())
     tee = kwds.pop('tee', False)
     update_nl = kwds.pop('update_nl', False)
     profile_time = kwds.pop('profile_time', False)

     # D_bar obj
     obj = 0.0
     for t in var_est_object._meas_times:
         for l in var_est_object._meas_lambdas:
             D_bar = sum(var_est_object.model.S[l, k].value*var_est_object.C_model.C[t, k] for k in var_est_object.component_set)
             obj += (var_est_object.model.D[t, l]-D_bar)**2
     
     var_est_object.C_model.objective = Objective(expr=obj)
             
     if profile_time:
         print('-----------------Solve_C--------------------')

     opt = SolverFactory(solver)

     for key, val in solver_opts.items():
         opt.options[key]=val
         
     solver_results = opt.solve(var_est_object.C_model,
                                logfile=var_est_object._tmp4,
                                tee=tee,
                                #keepfiles=True,
                                report_timing=profile_time)

     var_est_object.C_model.del_component('objective')
     
     for t in var_est_object._allmeas_times:
         for c in var_est_object.component_set:
             var_est_object.model.C[t, c].value = var_est_object.C_model.C[t, c].value

def solve_S(var_est_object, solver, **kwds):
    """Solves formulation 23 from Weifengs procedure with ipopt

       This method is not intended to be used by users directly

    Args:

    Returns:
        None

    """
    solver_opts = kwds.pop('solver_opts', dict())
    tee = kwds.pop('tee', False)
    update_nl = kwds.pop('update_nl', False)
    profile_time = kwds.pop('profile_time', False)
        
    for l in var_est_object._meas_lambdas:
        for c in var_est_object.component_set:
            var_est_object.S_model.S[l, c].value = var_est_object.model.S[l, c].value
            if hasattr(var_est_object.model, 'known_absorbance'):
                if c in var_est_object.model.known_absorbance:
                    if var_est_object.model.S[l, c].value != var_est_object.model.known_absorbance_data[c][l]:
                        var_est_object.model.S[l, c].set_value(var_est_object.model.known_absorbance_data[c][l])
                        var_est_object.S_model.S[l, c].fix()
    
    #D_bar obj
    obj = 0.0
    for t in var_est_object._meas_times:
        for l in var_est_object._meas_lambdas:
            D_bar = sum(var_est_object.S_model.S[l, k] * var_est_object.model.Z[t, k].value for k in var_est_object.component_set)
            obj += (D_bar - var_est_object.model.D[t, l]) ** 2
                
    var_est_object.S_model.objective = Objective(expr=obj)

    if profile_time:
        print('-----------------Solve_S--------------------')

    opt = SolverFactory(solver)

    for key, val in solver_opts.items():
        opt.options[key]=val
    solver_results = opt.solve(var_est_object.S_model,
                               logfile=var_est_object._tmp3,
                               tee=tee,
                               #keepfiles=True,
                               #show_section_timing=True,
                               report_timing=profile_time)

    var_est_object.S_model.del_component('objective')
    
    for l in var_est_object._meas_lambdas:
        for c in var_est_object.component_set:
            var_est_object.model.S[l, c].value = var_est_object.S_model.S[l, c].value
            if hasattr(var_est_object.model, 'known_absorbance'):
                if c in var_est_object.model.known_absorbance:
                    if var_est_object.model.S[l, c].value != var_est_object.model.known_absorbance_data[c][l]:
                        var_est_object.model.S[l, c].set_value(var_est_object.model.known_absorbance_data[c][l])
                        var_est_object.S_model.S[l, c].fix()
               

def solve_Z(var_est_object, solver, **kwds):
    """Solves formulation 20 in weifengs paper

       This method is not intended to be used by users directly

    Args:
    
        solver_opts (dict, optional): options passed to the nonlinear solver

        tee (bool,optional): flag to tell the optimizer whether to stream output
        to the terminal or not
    
        profile_time (bool,optional): flag to tell pyomo to time the construction and solution of the model. 
        Default False
    
        subset_lambdas (array_like,optional): Set of wavelengths to used in initialization problem 
        (Weifeng paper). Default all wavelengths.

    Returns:

        None

    """
    solver_opts = kwds.pop('solver_opts', dict())
    tee = kwds.pop('tee', False)
    profile_time = kwds.pop('profile_time', False)
    
    for t in var_est_object._allmeas_times:
        for k in var_est_object._sublist_components:
            if hasattr(var_est_object.model, 'non_absorbing'):
                if k not in var_est_object.model.non_absorbing:
                    pass
                else:
                    var_est_object.model.C[t, k].fixed = True

    obj = 0.0
    
    # Conc obj no sigma - will this work with missing data?
    for k in var_est_object._sublist_components:
        obj += sum((var_est_object.model.C[t, k]-var_est_object.model.Z[t, k])**2 for t in var_est_object._meas_times)
        

    var_est_object.model.z_objective = Objective(expr=obj)
    if profile_time:
        print('-----------------Solve_Z--------------------')
        
    opt = SolverFactory(solver)

    for key, val in solver_opts.items():
        opt.options[key]=val

    from pyomo.opt import ProblemFormat
    solver_results = opt.solve(var_est_object.model,
                               logfile=var_est_object._tmp2,
                               tee=tee,
                               #show_section_timing=True,
                               report_timing=profile_time)

    var_est_object.model.del_component('z_objective')

def solve_S_from_DC(var_est_object, C_dataFrame, tee=False, with_bounds=False, max_iter=200):
    """Solves a basic least squares problems with SVD.
    
    Args:
        C_dataFrame (DataFrame) data frame with concentration values
    
    Returns:
        DataFrame with estimated S_values 

    """
    D_data = var_est_object.model.D
    if var_est_object._n_meas_lambdas:
        # build Dij vector
        D_vector = np.zeros(var_est_object._n_meas_times*var_est_object._n_meas_lambdas)
        
        row  = []
        col  = []
        data = []    
        for i,t in enumerate(var_est_object._meas_times):
            for j,l in enumerate(var_est_object._meas_lambdas):
                for k,c in enumerate(var_est_object._mixture_components):
                    row.append(i*var_est_object._n_meas_lambdas+j)
                    col.append(j*var_est_object._n_components+k)
                    data.append(C_dataFrame[c][t])
                D_vector[i*var_est_object._n_meas_lambdas+j] = D_data[t,l]    
            
                    
        Bd = scipy.sparse.coo_matrix((data, (row, col)),
                                     shape=(var_est_object._n_meas_times*var_est_object._n_meas_lambdas,
                                            var_est_object._n_components*var_est_object._n_meas_lambdas))

        if not with_bounds:
            if var_est_object._n_meas_times == var_est_object._n_components:
                s_array = scipy.sparse.linalg.spsolve(Bd, D_vector)
            elif var_est_object._n_meas_times>var_est_object._n_components:
                result_ls = scipy.sparse.linalg.lsqr(Bd, D_vector,show=tee)
                s_array = result_ls[0]
            else:
                raise RuntimeError('Need n_t_meas >= var_est_object._n_components')
        else:
            nl = var_est_object._n_meas_lambdas
            nt = var_est_object._n_meas_times
            nc = var_est_object._n_components
            x0 = np.zeros(nl*nc)+1e-2
            M = Bd.tocsr()
            
            def F(x,M,rhs):
                return  rhs-M.dot(x)

            def JF(x,M,rhs):
                return -M

            if tee == True:
                verbose = 2
            else:
                verbose = 0
                
            res_lsq = least_squares(F,x0,JF,
                                    bounds=(0.0,np.inf),
                                    max_nfev=max_iter,
                                    verbose=verbose,args=(M,D_vector))
            s_array = res_lsq.x
            
        s_shaped = s_array.reshape((var_est_object._n_meas_lambdas,var_est_object._n_components))
    else:
        s_shaped = np.empty((var_est_object._n_meas_lambdas,var_est_object._n_components))

    return s_shaped


"""
Parameter handling
"""
from kipet.top_level.variable_names import VariableNames

__var = VariableNames()
parameter_set_name = __var.model_parameter

def calculate_parameter_averages(model_dict):
    
    p_dict = {}
    c_dict = {}
    
    for key, model in model_dict.items():
        for param, obj in getattr(model, __var.model_parameter).items():
            if param not in p_dict:
                p_dict[param] = obj.value
                c_dict[param] = 1
            else:
                p_dict[param] += obj.value
                c_dict[param] += 1
                
    avg_param = {param: p_dict[param]/c_dict[param] for param in p_dict.keys()}
    
    return avg_param
    
def initialize_parameters(model_dict):
    
    avg_param = calculate_parameter_averages(model_dict)
    
    for key, model in model_dict.items():
        for param, obj in getattr(model, __var.model_parameter).items():
            obj.value = avg_param[param] 
            
    return None

def check_initial_parameter_values(model_object):
    """Checks the initial parameter values and alters them if they violate the
    bounds.
    
    Args:
        model_object (pyomo ConcreteModel): A pyomo model instance of the current
            problem 
            
    Returns:
        None
        
    """
    
    for k, v in getattr(model_object, parameter_set_name).items():
        
        bound_push = 1.05
        
        if v.value >= v.ub:
            v.set_value(param.ub/bound_push)
        
        if v.value <= v.lb:
            v.set_value(v.lb*bound_push)
            
    return None
        
def set_scaled_parameter_bounds(model_object, parameter_set=None, rho=10, scaled=True, original_bounds=None):
    """Set the parameter values (scaled) for a given set of parameters
    
    Args:
        model_object (pyomo ConcreteModel): A pyomo model instance of the current
            problem 
            
        parameter_set (list): list of parameters to be considered, if None all
            parameters in the model are considered
            
        rho (float): ratio for setting the upper and lower bounds
        
        scaled (bool): True if the parameters are scaled
        
        original_bounds (bool): True if the original parameter bounds are to be
            used (instead of rho)
            
    Returns:
        None
    
    """
    param_set = param = getattr(model_object, parameter_set_name)
    
    if parameter_set is None:
        parameter_set = [p for p in param_set]
    
    for k, v in param_set.items():
        
        if k in parameter_set:
            
            if not scaled:
                ub = rho*v.value
                lb = 1/rho*v.value
            else:
                ub = rho
                lb = 1/rho
            
            if original_bounds is not None:
                if ub > original_bounds[k][1]:
                    ub = original_bounds[k][1]
                if lb < original_bounds[k][0]:
                    lb = original_bounds[k][0]
                
            v.setlb(lb)
            v.setub(ub)
            v.unfix()
            
            if scaled:
                v.set_value(1)
    
        else:
            v.fix()

    return None


"""
Initialization for method from Chen et al. 2016 
"""
import os

from pyomo.environ import (
    Objective,
    SolverFactory,
    )

from kipet.core_methods.ResultsObject import *
from kipet.variance_methods.chen_method_pyomo import *
from kipet.variance_methods.chen_method_scipy import *

# Try to replace var_est_object with var_est_object

def run_method(var_est_object, solver, run_opt_kwargs):
    """This is the original method for estimating variances from Chen et
    al. 2016
    
    """
    solver_opts = run_opt_kwargs.pop('solver_opts', dict())
    max_iter = run_opt_kwargs.pop('max_iter', 400)
    tol = run_opt_kwargs.pop('tolerance', 5.0e-5)
    tee = run_opt_kwargs.pop('tee', False)
    norm_order = run_opt_kwargs.pop('norm',np.inf)
    A = run_opt_kwargs.pop('subset_lambdas', None)
    init_C = run_opt_kwargs.pop('init_C', None)
    lsq_ipopt = run_opt_kwargs.pop('lsq_ipopt', False)
    species_list = run_opt_kwargs.pop('subset_components', None)
    fixed_device_var = run_opt_kwargs.pop('fixed_device_variance', None)
    
    if init_C is None:

        solve_initalization(var_est_object, solver, 
                                  subset_lambdas=A, 
                                  solver_opts=solver_opts, 
                                  tee=tee)
    else:
        for t in var_est_object._allmeas_times:
            for k in var_est_object._mixture_components:
                var_est_object.model.C[t, k].value = init_C[k][t]
                var_est_object.model.Z[t, k].value = init_C[k][t]

        # This comes from Optimizer
        s_array = _solve_S_from_DC(var_est_object, init_C)
        S_frame = pd.DataFrame(data=s_array,
                               columns=var_est_object._mixture_components,
                               index=var_est_object._meas_lambdas)
                                                                        
        if hasattr(var_est_object, '_abs_components'):
            component_set = var_est_object._abs_components
        else:
            component_set = var_est_object._mixture_components
            
        for l in var_est_object._meas_lambdas:
            for k in component_set:
                var_est_object.model.S[l, k].value = S_frame[k][l]  # 1e-2
                if hasattr(var_est_object.model, 'known_absorbance'):
                    if k in var_est_object.model.known_absorbance:
                        var_est_object.model.S[l, k].value = var_est_object.model.known_absorbance_data[k][l]
        
        
    print("{: >11} {: >20}".format('Iter', '|Zi-Zi+1|'))
    logiterfile = "iterations.log"
    if os.path.isfile(logiterfile):
        os.remove(logiterfile)

    if lsq_ipopt:
        build_s_model(var_est_object)
        build_c_model(var_est_object)
    else:
        if species_list is None:
            build_scipy_lsq_arrays(var_est_object)
        else:
            lsq_ipopt = True
            build_s_model(var_est_object)
            build_c_model(var_est_object)
        
    for it in range(max_iter):
        
        rb = ResultsObject()
        
        vars_to_load = ['Z', 'C', 'Cs', 'S', 'Y']
        if not hasattr(var_est_object, '_abs_components'):
            vars_to_load.remove('Cs')
        rb.load_from_pyomo_model(var_est_object.model, to_load=vars_to_load)
        
        solve_Z(var_est_object, solver)

        if lsq_ipopt:
            solve_S(var_est_object, solver)
            solve_C(var_est_object, solver)
        else:
            solved_s = solve_s_scipy(var_est_object)
            solved_c = solve_c_scipy(var_est_object)
            
        ra = ResultsObject()    
        vars_to_load = ['Z', 'C', 'Cs', 'S']
        if not hasattr(var_est_object, '_abs_components'):
            vars_to_load.remove('Cs')
        ra.load_from_pyomo_model(var_est_object.model, to_load=vars_to_load)
        
        r_diff = compute_diff_results(rb,ra)
        Z_norm = r_diff.compute_var_norm('Z', norm_order)
        
        if it > 0:
            print("{: >11} {: >20}".format(it, Z_norm))
        _log_iterations(var_est_object, logiterfile, it)
        if Z_norm < tol and it >= 1:
            break
            
    results = ResultsObject()
    
    vars_to_load = ['Z', 'dZdt', 'X', 'dXdt', 'C', 'Cs', 'S', 'Y']
    if not hasattr(var_est_object, '_abs_components'):
        vars_to_load.remove('Cs')
    results.load_from_pyomo_model(var_est_object.model, to_load=vars_to_load)

    print('Iterative optimization converged. Estimating variances now')
    solved_variances = _solve_variances(var_est_object,
                                        results, 
                                        fixed_dev_var=fixed_device_var)
    
    compute_D_given_SC(var_est_object, results)
    
    results.P = {name: var_est_object.model.P[name].value for name in var_est_object.model.parameter_names}

    # removes temporary files. This needs to be changes to work with pyutilib
    if os.path.exists(var_est_object._tmp2):
        os.remove(var_est_object._tmp2)
    if os.path.exists(var_est_object._tmp3):
        os.remove(var_est_object._tmp3)
    if os.path.exists(var_est_object._tmp4):
        os.remove(var_est_object._tmp4)
        
    return results

def solve_initalization(var_est_object, solver, **kwds):
    """Solves formulation 19 in weifengs paper

       This method is not intended to be used by users directly

    Args:
        sigma_sq (dict): variances 
    
        tee (bool,optional): flag to tell the optimizer whether to stream output
        to the terminal or not
    
        profile_time (bool,optional): flag to tell pyomo to time the construction and solution of the model. 
        Default False
    
        subset_lambdas (array_like,optional): Set of wavelengths to used in initialization problem 
        (Weifeng paper). Default all wavelengths.

    Returns:

        None

    """
    solver_opts = kwds.pop('solver_opts', dict())
    tee = kwds.pop('tee', True)
    set_A = kwds.pop('subset_lambdas', list())
    profile_time = kwds.pop('profile_time', False)
    sigmas_sq = kwds.pop('variances', dict())

    if not set_A:
        set_A = var_est_object._meas_lambdas
    
    keys = sigmas_sq.keys()
    # added due to new structure for non_abs species, non-absorbing species not included in S and Cs as subset of C (CS):
    
    for k in var_est_object.component_set:
        if k not in keys:
            sigmas_sq[k] = 0.0

    print("Solving Initialization Problem\n")

    # Check this!!!!
    var_est_object._warn_if_D_negative()        
    
    obj = 0.0
    for t in var_est_object._meas_times:
        for l in set_A:
            D_bar = sum(var_est_object.model.Z[t, k] * var_est_object.model.S[l, k] for k in var_est_object.component_set)
            obj += (var_est_object.model.D[t, l] - D_bar) ** 2
                
    var_est_object.model.init_objective = Objective(expr=obj)
    
    opt = SolverFactory(solver)

    for key, val in solver_opts.items():
        opt.options[key]=val
    solver_results = opt.solve(var_est_object.model,
                               tee=tee,
                               report_timing=profile_time)

    for t in var_est_object._allmeas_times:
        for k in var_est_object._mixture_components:
            if k in sigmas_sq and sigmas_sq[k] > 0.0:
                var_est_object.model.C[t, k].value = np.random.normal(var_est_object.model.Z[t, k].value, sigmas_sq[k])
            else:
                var_est_object.model.C[t, k].value = var_est_object.model.Z[t, k].value
            
    var_est_object.model.del_component('init_objective')
    
    return None

def _log_iterations(var_est_object, filename, iteration):
    """log solution of each subproblem in Weifengs procedure

       This method is not intended to be used by users directly

    Args:

    Returns:
        None

    """
    with open(filename, "a") as f:
        f.write("\n#######################Iteration {}#######################\n".format(iteration))
        with open(var_est_object._tmp2,'r') as tf:
            f.write("\n#######################Solve Z {}#######################\n".format(iteration))
            f.write(tf.read())
        with open(var_est_object._tmp3,'r') as tf:
            f.write("\n#######################Solve S {}#######################\n".format(iteration))
            f.write(tf.read())
        with open(var_est_object._tmp4,'r') as tf:
            f.write("\n#######################Solve C {}#######################\n".format(iteration))
            f.write(tf.read())

    return None

def _solve_variances(var_est_object, results, fixed_dev_var = None):
    """Solves formulation 23 in weifengs paper (using scipy least_squares)

       This method is not intended to be used by users directly

    Args:
        results (ResultsObject): Data obtained from Weifengs procedure
        fixed_device_var(float, optional): if we have device variance we input it here

    Returns:
        bool indicated if variances were estimated succesfully.

    """
    
    nl = var_est_object._n_meas_lambdas
    nt = var_est_object._n_meas_times
    b = np.zeros((nl, 1))

    variance_dict = dict()
    n_val = len(var_est_object.component_set)
        
    A = np.ones((nl, n_val + 1))
    reciprocal_nt = 1.0/nt
    for i, l in enumerate(var_est_object._meas_lambdas):
        for j, t in enumerate(var_est_object._meas_times):
            D_bar = 0.0
            for w, k in enumerate(var_est_object.component_set):
                A[i, w] = results.S[k][l]**2
                D_bar += results.S[k][l]*results.Z[k][t]
            b[i] += (var_est_object.model.D[t, l] - D_bar)**2
        b[i] *= reciprocal_nt

    if fixed_dev_var == None:
        res_lsq = np.linalg.lstsq(A, b, rcond=None)
        all_nonnegative = True
        n_vars = n_val + 1

        for i in range(n_vars):
            if res_lsq[0][i] < 0.0:
                if res_lsq[0][i] < -1e-5:
                    all_nonnegative = False
                else:
                    res_lsq[0][i] = abs(res_lsq[0][i])
            res_lsq[0][i]

        if not all_nonnegative:
            x0 = np.zeros(n_val + 1) + 1e-2
            bb = np.zeros(nl)
            for i in range(nl):
                bb[i] = b[i]

            def F(x, M, rhs):
                return rhs - M.dot(x)

            def JF(x, M, rhs):
                return -M

            res_lsq = least_squares(F, x0, JF,
                                    bounds=(0.0, np.inf),
                                    verbose=2, args=(A, bb))
            for i, k in enumerate(var_est_object.component_set):
                variance_dict[k] = res_lsq.x[i]
            variance_dict['device'] = res_lsq.x[n_val]
            results.sigma_sq = variance_dict
            return res_lsq.success

        else:
            for i, k in enumerate(var_est_object.component_set):
                variance_dict[k] = res_lsq[0][i][0]
            variance_dict['device'] = res_lsq[0][n_val][0]
            results.sigma_sq = variance_dict

    if fixed_dev_var:
        bp = np.zeros((nl, 1))
        for i, l in enumerate(var_est_object._meas_lambdas):
            bp[i] = b[i] - fixed_dev_var
        Ap = np.zeros((nl, n_val))
        for i, l in enumerate(var_est_object._meas_lambdas):
            for j, t in enumerate(var_est_object._meas_times):
                for w, k in enumerate(var_est_object.component_set):
                    Ap[i, w] = results.S[k][l] ** 2

        res_lsq = np.linalg.lstsq(Ap, bp, rcond=None)
        all_nonnegative = True
        n_vars = n_val
        for i in range(n_vars):
            if res_lsq[0][i] < 0.0:
                if res_lsq[0][i] < -1e-5:
                    all_nonnegative = False
                else:
                    res_lsq[0][i] = abs(res_lsq[0][i])
            res_lsq[0][i]

        for i, k in enumerate(var_est_object.component_set):
            variance_dict[k] = res_lsq[0][i][0]
        variance_dict['device'] = fixed_dev_var
        results.sigma_sq = variance_dict
    
    return 1

def build_scipy_lsq_arrays(var_est_object):
    """Creates arrays for scipy solvers

       This method is not intended to be used by users directly

    Args:

    Returns:
        None

    """
    var_est_object._d_array = np.zeros((var_est_object._n_meas_times, var_est_object._n_meas_lambdas))
    for i, t in enumerate(var_est_object._meas_times):
        for j, l in enumerate(var_est_object._meas_lambdas):
            var_est_object._d_array[i,j] = var_est_object.model.D[t,l]

    if hasattr(var_est_object,'_abs_components'):
        n_val = var_est_object._nabs_components
    else:
        n_val = var_est_object._n_components
        
    var_est_object._s_array = np.ones(var_est_object._n_meas_lambdas * n_val) 
    var_est_object._z_array = np.ones(var_est_object._n_allmeas_times * n_val)
    var_est_object._c_array = np.ones(var_est_object._n_allmeas_times * n_val)
 
    return None

def compute_diff_results(results1,results2):
    diff_results = ResultsObject()
    diff_results.Z = results1.Z - results2.Z
    diff_results.S = results1.S - results2.S
    diff_results.C = results1.C - results2.C
    return diff_results

def compute_D_given_SC(var_est_object, results,sigma_d=0):
        # this requires results to have S and C computed already
    d_results = []

    if hasattr(var_est_object, '_abs_components'):  # added for removing non_abs ones from first term in obj CS
        for i, t in enumerate(var_est_object._allmeas_times):
            if t in var_est_object._meas_times:
                for j, l in enumerate(var_est_object._meas_lambdas):
                    suma = 0.0
                    for w, k in enumerate(var_est_object._abs_components):
                        Cs = results.Cs[k][t]  # just the absorbing ones
                        Ss = results.S[k][l]
                        suma += Cs * Ss
                    if sigma_d:
                        suma += np.random.normal(0.0, sigma_d)
                    d_results.append(suma)

    else:
        for i, t in enumerate(var_est_object._allmeas_times):
            if t in var_est_object._meas_times:
                for j, l in enumerate(var_est_object._meas_lambdas):
                    suma = 0.0
                    for w, k in enumerate(var_est_object._mixture_components):
                        C = results.C[k][t]
                        S = results.S[k][l]
                        suma += C * S
                    if sigma_d:
                        suma += np.random.normal(0.0, sigma_d)
                    d_results.append(suma)

    d_array = np.array(d_results).reshape((var_est_object._n_meas_times, var_est_object._n_meas_lambdas))
    results.D = pd.DataFrame(data=d_array,
                             columns=var_est_object._meas_lambdas,
                             index=var_est_object._meas_times)



"""
This module contains the code related to jumps - removes redundancy
"""
from pyomo.core import *
from pyomo.dae import *
from pyomo.environ import *

from kipet.common.VisitorClasses import ReplacementVisitor

class JumpsMixin():
        
    def set_up_jumps(self, kwargs):
    
        var_dic = kwargs.pop("jump_states", None)
        jump_times = kwargs.pop("jump_times", None)
        feed_times = kwargs.pop("feed_times", None)
    
        self.disc_jump_v_dict = var_dic
        self.jump_times_dict = jump_times  # now dictionary
        self.feed_times_set = feed_times
        if not isinstance(self.disc_jump_v_dict, dict):
            print("disc_jump_v_dict is of type {}".format(type(self.disc_jump_v_dict)))
            raise Exception  # wrong type
        if not isinstance(self.jump_times_dict, dict):
            print("disc_jump_times is of type {}".format(type(self.jump_times_dict)))
            raise Exception  # wrong type
        count = 0
        for i in self.jump_times_dict.keys():
            for j in self.jump_times_dict[i].items():
                count += 1
        if len(self.feed_times_set) > count:
            raise Exception("Error: Check feed time points in set feed_times and in jump_times again.\n"
                        "There are more time points in feed_times than jump_times provided.")
        self.load_discrete_jump()
    
    def load_discrete_jump(self):
        self.jump = True
    
        zeit = None
        for i in self.model.component_objects(ContinuousSet):
            zeit = i
            break
        if zeit is None:
            raise Exception('No continuous_set')
    
        self.time_set = zeit.name
    
        tgt_cts = getattr(self.model, self.time_set)  ## please correct me (not necessary!)
        self.ncp = tgt_cts.get_discretization_info()['ncp']
        fe_l = tgt_cts.get_finite_elements()
        fe_list = [fe_l[i + 1] - fe_l[i] for i in range(0, len(fe_l) - 1)]
    
        for i in range(0, len(fe_list)):  # test whether integer elements
            self.jump_constraints(i)
      
    # I want to change this spaghetti
    def jump_constraints(self, fe):
        # type: (int) -> None
        """ Take the current state of variables of the initializing model at fe and load it into the tgt_model
        Note that this will skip fixed variables as a safeguard.
    
        Args:
            fe (int): The current finite element to be patched (tgt_model).
        """
        ###########################
        if not isinstance(fe, int):
            raise Exception  # wrong type
        ttgt = getattr(self.model, self.time_set)
        ##############################
        # Inclusion of discrete jumps: (CS)
        if self.jump:
            vs = ReplacementVisitor()  #: trick to replace variables
            kn = 0
            for ki in self.jump_times_dict.keys():
                if not isinstance(ki, str):
                    print("ki is not str")
                vtjumpkeydict = self.jump_times_dict[ki]
                for l in vtjumpkeydict.keys():
                    self.jump_time = vtjumpkeydict[l]
                    # print('jumptime:',self.jump_time)
                    self.jump_fe, self.jump_cp = self.fe_cp(ttgt, self.jump_time)
                    if self.jump_time not in self.feed_times_set:
                        raise Exception("Error: Check feed time points in set feed_times and in jump_times again.\n"
                                        "They do not match.\n"
                                        "Jump_time is not included in feed_times.")
                    # print('jump_el, el:',self.jump_fe, fe)
                    if fe == self.jump_fe + 1:
                        #################################
                        for v in self.disc_jump_v_dict.keys():
                            if not isinstance(v, str):
                                print("v is not str")
                            vkeydict = self.disc_jump_v_dict[v]
                            for k in vkeydict.keys():
                                if k == l:  # Match in between two components of dictionaries
                                    var = getattr(self.model, v)
                                    dvar = getattr(self.model, "d" + v + "dt")
                                    con_name = 'd' + v + 'dt_disc_eq'
                                    con = getattr(self.model, con_name)
    
                                    self.model.add_component(v + "_dummy_eq_" + str(kn), ConstraintList())
                                    conlist = getattr(self.model, v + "_dummy_eq_" + str(kn))
                                    varname = v + "_dummy_" + str(kn)
                                    self.model.add_component(varname, Var([0]))  #: this is now indexed [0]
                                    vdummy = getattr(self.model, varname)
                                    vs.change_replacement(vdummy[0])   #: who is replacing.
                                    # self.model.add_component(varname, Var())
                                    # vdummy = getattr(self.model, varname)
                                    jump_delta = vkeydict[k]
                                    self.model.add_component(v + '_jumpdelta' + str(kn),
                                                             Param(initialize=jump_delta))
                                    jump_param = getattr(self.model, v + '_jumpdelta' + str(kn))
                                    if not isinstance(k, tuple):
                                        k = (k,)
                                    exprjump = vdummy[0] - var[(self.jump_time,) + k] == jump_param  #: this cha
                                    # exprjump = vdummy - var[(self.jump_time,) + k] == jump_param
                                    self.model.add_component("jumpdelta_expr" + str(kn), Constraint(expr=exprjump))
                                    for kcp in range(1, self.ncp + 1):
                                        curr_time = self.t_ij(ttgt, self.jump_fe + 1, kcp)
                                        if not isinstance(k, tuple):
                                            knew = (k,)
                                        else:
                                            knew = k
                                        idx = (curr_time,) + knew
                                        con[idx].deactivate()
                                        e = con[idx].expr
                                        suspect_var = e.args[0].args[1].args[0].args[0].args[1]  #: seems that
                                        # e = con[idx].expr.clone()
                                        # e.args[0].args[1] = vdummy
                                        # con[idx].set_value(e)
                                        vs.change_suspect(id(suspect_var))  #: who to replace
                                        e_new = vs.dfs_postorder_stack(e)  #: replace
                                        con[idx].set_value(e_new)
                                        conlist.add(con[idx].expr)
                    kn = kn + 1
    
    @staticmethod
    def t_ij(time_set, i, j):
        # type: (ContinuousSet, int, int) -> float
        """Return the corresponding time(continuous set) based on the i-th finite element and j-th collocation point
        From the NMPC_MHE framework by @dthierry.
    
        Args:
            time_set (ContinuousSet): Parent Continuous set
            i (int): finite element
            j (int): collocation point
    
        Returns:
            float: Corresponding index of the ContinuousSet
        """
        if i < time_set.get_discretization_info()['nfe']:
            h = time_set.get_finite_elements()[i + 1] - time_set.get_finite_elements()[i]  #: This would work even for 1 fe
        else:
            h = time_set.get_finite_elements()[i] - time_set.get_finite_elements()[i - 1]  #: This would work even for 1 fe
        tau = time_set.get_discretization_info()['tau_points']
        fe = time_set.get_finite_elements()[i]
        time = fe + tau[j] * h
        return round(time, 6)
    
    @staticmethod
    def fe_cp(time_set, feedtime):
        """Return the corresponding fe and cp for a given time
         Args:
            time_set:
            t:
        
        type: (ContinuousSet, float) -> tuple
                 
        """
        fe_l = time_set.get_lower_element_boundary(feedtime)
        fe = None
        j = 0
        for i in time_set.get_finite_elements():
            if fe_l == i:
                fe = j
                break
            j += 1
        h = time_set.get_finite_elements()[1] - time_set.get_finite_elements()[0]
        tauh = [i * h for i in time_set.get_discretization_info()['tau_points']]
        j = 0  #: Watch out for LEGENDRE
        cp = None
        for i in tauh:
            if round(i + fe_l, 6) == feedtime:
                cp = j
                break
            j += 1
        return fe, cp